#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cengal.code_flow_control.gc import DisableGC
from cengal.modules_management.alternative_import import alt_import
from cengal.time_management.repeat_for_a_time import Tracer
from cengal.time_management.cpu_clock_cycles import perf_counter
import enum
from functools import lru_cache
from typing import Set, Union, Optional, Any, Dict, Tuple, List


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class Tags(enum.Enum):
    tested = 0
    deep = 1  # can work with nested data structures
    superficial = 2  # can work with only single-layer of data
    multi_platform = 3  # serialized data can be deserialized by other languages/interpreters
    current_platform = 4  # serialized data can be deserialized by current type of python interpreter (only by CPython or only by PyPy for example)
    multicast = 5  # ?
    unique = 6  # ?
    can_use_set = 7  # can serialize set type
    can_use_bytes = 8  # can serialize bytes type
    can_use_custom_types = 9  # can serialize custom types
    can_use_lambda_functions = 10
    decode_str_as_str = 11  # does not convert strings to bytes
    decode_bytes_as_bytes = 12  # does not convert bytes to strings
    decode_str_and_bytes_as_requested = 13  # can chose: convert all str/bytes to str or to convert all str/bytes to bytes.
    decode_list_and_tuple_as_requested = 14  # can chose: convert all list/tuple to list or to convert all list/tuple to tuple.
    decode_tuple_as_tuple = 15  # does not conver tuples to list/set
    decode_list_as_list = 16  # does not convert lists to tuple/set
    explicit_format_version = 17  # Example: pickle can be of a different versions since python-version depent so it is Not tagged as an explicit_format_version.
    fast = 17
    compat = 18
    compat_with_python_below_3_8 = 19
    compat_with_python_abowe_3_8 = 20
    decode_set_as_set = 21  # does not convert sets to tuple/list


class DataFormats(enum.Enum):
    any = 0  # we don't care about the serialized data format
    json = 1  # serialized data is json
    binary = 2  # serialized data is binary (not really human-readable)
    text = 3  # serialized data is human-readable
    yaml = 4
    toml = 5
    messagepack = 6


SerializerFeatures = Set[Union[Tags, DataFormats]]
SerializerFeaturesTuple = Tuple[Union[Tags, DataFormats]]


# !!! Keep an order! Add new serializers to the end of the list only !!!
class Serializers(enum.Enum):
    json = 0
    simplejson = 1
    ujson = 2
    orjson = 3
    tnetstring = 4
    pynetstring = 5
    msgpack_fast = 6
    msgpack = 7
    cbor = 8
    cbor2 = 9
    marshal = 10
    marshal_compat_4 = 11
    marshal_compat_3 = 12
    marshal_compat_2 = 13
    marshal_compat_1 = 14
    marshal_compat_0 = 15
    pickle = 16
    pickle_default = 17
    pickle_compat_5 = 18
    pickle_compat_4 = 19
    pickle_compat_3 = 20
    pickle_compat_2 = 21
    pickle_compat_1 = 22
    cpickle = 23
    cpickle_default = 24
    cpickle_compat_5 = 25
    cpickle_compat_4 = 26
    cpickle_compat_3 = 27
    cpickle_compat_2 = 28
    cpickle_compat_1 = 29
    cloudpickle = 30
    cloudpickle_compat_5 = 31
    cloudpickle_compat_4 = 32
    cloudpickle_compat_3 = 33
    cloudpickle_compat_2 = 34
    cloudpickle_compat_1 = 35
    msgspec_json = 36
    msgspec_messagepack = 37
    msgspec_yaml = 38
    msgspec_toml = 39


import datetime
import decimal
import fractions
import re
from email.message import Message as emailMessage
import uuid
import ipaddress
from types import CodeType
none_type = type(None)
SUPPORTED_TYPES = {
    Serializers.json: {int, float, bool, none_type, str, list, dict},
    Serializers.simplejson: {int, float, bool, none_type, str, list, dict},
    Serializers.ujson: {int, float, bool, none_type, str, list, dict},
    Serializers.orjson: {int, float, bool, none_type, str, list, dict},
    Serializers.tnetstring: {int, float, bool, none_type, str, list, dict},
    Serializers.pynetstring: {int, float, bool, none_type, str, list, dict},
    Serializers.msgpack: {int, float, bool, none_type, str, bytes, list, dict},
    Serializers.msgpack_fast: {int, float, bool, none_type, str, bytes, tuple, dict},
    Serializers.cbor: {int, float, bool, none_type, str, list, dict},
    Serializers.cbor2: {
        int, float, bool, none_type, str, list, dict,
        datetime.datetime, datetime.date, decimal.Decimal, fractions.Fraction,
        re.Pattern, emailMessage, uuid.UUID, set, 
        ipaddress.IPv4Address, ipaddress.IPv6Address, ipaddress.IPv4Network, ipaddress.IPv6Network,
    },
    Serializers.marshal: {
        bool, int, float, complex, str, bytes, bytearray, tuple, list, set, frozenset, dict,
        CodeType, none_type, Ellipsis, StopIteration
    },
    Serializers.pickle: {},
    Serializers.cpickle: {},
    Serializers.cloudpickle: {},
    Serializers.msgspec_messagepack: {int, float, bool, none_type, str, bytes, list, dict},
    Serializers.msgspec_json: {int, float, bool, none_type, str, list, dict},
    Serializers.msgspec_yaml: {
        int, float, bool, none_type, str, list, dict,
        datetime.datetime, datetime.date,
    },
    # Serializers.msgspec_toml: {
    #     int, float, bool, none_type, str, list, dict,
    #     datetime.datetime, datetime.date, datetime.time,
    # },
    Serializers.msgspec_toml: {
        str, dict,
        datetime.datetime, datetime.date, datetime.time,
    },
}


SERIALIZERS_DESCRIPTION = {
    Serializers.json: ('json', {
        Tags.deep,
        Tags.tested,
        Tags.superficial,
        Tags.multi_platform,
        Tags.current_platform,
        Tags.decode_str_as_str,
        Tags.decode_list_as_list,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.json,
        DataFormats.text,
    }),
    Serializers.simplejson: ('simplejson', {
        Tags.deep,
        Tags.tested,
        Tags.superficial,
        Tags.multi_platform,
        Tags.current_platform,
        Tags.can_use_bytes,
        Tags.decode_str_as_str,
        Tags.decode_list_as_list,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.json,
        DataFormats.text,
    }),
    Serializers.ujson: ('ujson', {
        Tags.deep,
        Tags.superficial,
        Tags.multi_platform,
        Tags.current_platform,
        Tags.can_use_bytes,
        Tags.decode_str_as_str,
        Tags.decode_list_as_list,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.json,
        DataFormats.text,
    }),
    Serializers.orjson: ('orjson', {
        Tags.deep,
        Tags.superficial,
        Tags.multi_platform,
        Tags.current_platform,
        Tags.can_use_bytes,
        Tags.decode_str_as_str,
        Tags.decode_list_as_list,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.json,
        DataFormats.text,
    }),
    Serializers.tnetstring: ('tnetstring', {
        Tags.superficial,
        Tags.current_platform,
        Tags.decode_str_as_str,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.text,
    }),
    Serializers.pynetstring: ('pynetstring', {
        Tags.superficial,
        Tags.multi_platform,
        Tags.current_platform,
        Tags.decode_str_as_str,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.text,
    }),
    Serializers.msgpack: ('msgpack', {
        Tags.tested,
        Tags.deep,
        Tags.superficial,
        Tags.multi_platform,
        Tags.current_platform,
        Tags.can_use_bytes,
        Tags.decode_str_as_str,
        Tags.decode_bytes_as_bytes,
        Tags.decode_str_and_bytes_as_requested,
        Tags.decode_list_and_tuple_as_requested,
        Tags.decode_list_as_list,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.binary,
        DataFormats.messagepack,
    }),
    Serializers.msgpack_fast: ('msgpack_fast', {
        Tags.tested,
        Tags.deep,
        Tags.superficial,
        Tags.multi_platform,
        Tags.current_platform,
        Tags.can_use_bytes,
        Tags.decode_str_as_str,
        Tags.decode_bytes_as_bytes,
        Tags.decode_str_and_bytes_as_requested,
        Tags.decode_list_and_tuple_as_requested,
        Tags.decode_tuple_as_tuple,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.binary,
        DataFormats.messagepack,
    }),
    Serializers.cbor: ('cbor', {
        Tags.tested,
        Tags.deep,
        Tags.superficial,
        Tags.multi_platform,
        Tags.current_platform,
        Tags.can_use_bytes,
        Tags.decode_str_as_str,
        Tags.decode_bytes_as_bytes,
        Tags.decode_list_as_list,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.binary,
    }),
    Serializers.cbor2: ('cbor2', {
        Tags.tested,
        Tags.deep,
        Tags.superficial,
        Tags.multi_platform,
        Tags.current_platform,
        Tags.can_use_bytes,
        Tags.decode_str_as_str,
        Tags.decode_bytes_as_bytes,
        Tags.decode_list_as_list,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.binary,
    }),
    Serializers.marshal: ('marshal', {
        Tags.tested,
        Tags.deep,
        Tags.superficial,
        Tags.current_platform,
        Tags.can_use_bytes,
        Tags.decode_str_as_str,
        Tags.decode_bytes_as_bytes,
        Tags.can_use_set,
        Tags.decode_tuple_as_tuple,
        Tags.decode_list_as_list,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.binary,
    }),
    Serializers.pickle: ('pickle', {
        Tags.tested,
        Tags.deep,
        Tags.superficial,
        Tags.current_platform,
        Tags.can_use_bytes,
        Tags.can_use_set,
        Tags.can_use_custom_types,
        Tags.decode_str_as_str,
        Tags.decode_bytes_as_bytes,
        Tags.decode_tuple_as_tuple,
        Tags.decode_list_as_list,
        DataFormats.any,
        DataFormats.binary,
    }),
    Serializers.cpickle: ('cPickle', {
        Tags.deep,
        Tags.superficial,
        Tags.current_platform,
        Tags.can_use_bytes,
        Tags.can_use_set,
        Tags.can_use_custom_types,
        Tags.decode_str_as_str,
        Tags.decode_bytes_as_bytes,
        Tags.decode_tuple_as_tuple,
        Tags.decode_list_as_list,
        DataFormats.any,
        DataFormats.binary,
    }),
    Serializers.cloudpickle: ('cloudpickle', {
        Tags.deep,
        Tags.superficial,
        Tags.current_platform,
        Tags.can_use_bytes,
        Tags.can_use_set,
        Tags.can_use_custom_types,
        Tags.can_use_lambda_functions,
        Tags.decode_str_as_str,
        Tags.decode_bytes_as_bytes,
        Tags.decode_tuple_as_tuple,
        Tags.decode_list_as_list,
        DataFormats.any,
        DataFormats.binary,
    }),
    Serializers.msgspec_messagepack: ('msgspec_messagepack', {
        Tags.tested,
        Tags.deep,
        Tags.superficial,
        Tags.multi_platform,
        Tags.current_platform,
        Tags.can_use_set,
        Tags.can_use_bytes,
        Tags.decode_str_as_str,
        Tags.decode_bytes_as_bytes,
        Tags.decode_list_as_list,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.binary,
        DataFormats.messagepack,
    }),
    Serializers.msgspec_json: ('msgspec_json', {
        Tags.tested,
        Tags.deep,
        Tags.superficial,
        Tags.multi_platform,
        Tags.current_platform,
        Tags.can_use_set,
        Tags.decode_str_as_str,
        Tags.decode_list_as_list,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.json,
        DataFormats.text,
    }),
    Serializers.msgspec_yaml: ('msgspec_yaml', {
        Tags.tested,
        Tags.deep,
        Tags.superficial,
        Tags.multi_platform,
        Tags.current_platform,
        Tags.can_use_set,
        Tags.decode_str_as_str,
        Tags.decode_list_as_list,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.yaml,
        DataFormats.text,
    }),
    Serializers.msgspec_toml: ('msgspec_toml', {
        Tags.tested,
        Tags.deep,
        Tags.superficial,
        Tags.multi_platform,
        Tags.current_platform,
        Tags.can_use_set,
        Tags.decode_str_as_str,
        Tags.decode_list_as_list,
        Tags.explicit_format_version,
        DataFormats.any,
        DataFormats.toml,
        DataFormats.text,
    }),
}


ujson_dump_skip_parameters = {'skipkeys', 'check_circular', 'allow_nan', 'cls', 'separators', 'default'}
ujson_dumps_skip_parameters = {'skipkeys', 'check_circular', 'allow_nan', 'cls', 'separators', 'default'}
ujson_load_skip_parameters = {'cls', 'object_hook', 'parse_float', 'parse_int', 'parse_constant', 'object_pairs_hook'}
ujson_loads_skip_parameters = {'cls', 'object_hook', 'parse_float', 'parse_int', 'parse_constant', 'object_pairs_hook'}


orjson_dump_skip_parameters = {'skipkeys', 'check_circular', 'allow_nan', 'cls', 'separators', 'default'}
orjson_dumps_skip_parameters = {'skipkeys', 'check_circular', 'allow_nan', 'cls', 'separators', 'default'}
orjson_load_skip_parameters = {'cls', 'object_hook', 'parse_float', 'parse_int', 'parse_constant', 'object_pairs_hook'}
orjson_loads_skip_parameters = {'cls', 'object_hook', 'parse_float', 'parse_int', 'parse_constant', 'object_pairs_hook'}


EXISTING_SERIALIZERS = set()  # type: Set[Serializers]
with alt_import('json') as json:
    if json is not None:
        EXISTING_SERIALIZERS.add(Serializers.json)
with alt_import('simplejson') as simplejson:
    if simplejson is not None:
        EXISTING_SERIALIZERS.add(Serializers.simplejson)
with alt_import('ujson') as ujson:
    if ujson is not None:
        EXISTING_SERIALIZERS.add(Serializers.ujson)
with alt_import('orjson') as orjson:
    if orjson is not None:
        EXISTING_SERIALIZERS.add(Serializers.orjson)
with alt_import('tnetstring') as tnetstring:
    if tnetstring is not None:
        EXISTING_SERIALIZERS.add(Serializers.tnetstring)
# with alt_import('pynetstring') as pynetstring:
#     if pynetstring is not None:
#         EXISTING_SERIALIZERS.add(Serializers.pynetstring)
with alt_import('msgpack') as msgpack:
    if msgpack is not None:
        EXISTING_SERIALIZERS.add(Serializers.msgpack_fast)
        EXISTING_SERIALIZERS.add(Serializers.msgpack)
with alt_import('msgspec') as msgspec:
    if msgspec is not None:
        EXISTING_SERIALIZERS.add(Serializers.msgspec_messagepack)
        EXISTING_SERIALIZERS.add(Serializers.msgspec_json)
        EXISTING_SERIALIZERS.add(Serializers.msgspec_yaml)
        EXISTING_SERIALIZERS.add(Serializers.msgspec_toml)
with alt_import('cbor') as cbor:
    if cbor is not None:
        EXISTING_SERIALIZERS.add(Serializers.cbor)
with alt_import('cbor2') as cbor2:
    if cbor2 is not None:
        EXISTING_SERIALIZERS.add(Serializers.cbor2)
with alt_import('marshal') as marshal:
    if marshal is not None:
        EXISTING_SERIALIZERS.add(Serializers.marshal)
with alt_import('pickle') as pickle:
    if pickle is not None:
        EXISTING_SERIALIZERS.add(Serializers.pickle)
with alt_import('cPickle') as cPickle:
    if cPickle is not None:
        EXISTING_SERIALIZERS.add(Serializers.cpickle)
with alt_import('cloudpickle') as cloudpickle:
    if cloudpickle is not None:
        EXISTING_SERIALIZERS.add(Serializers.cloudpickle)


class Serializer:
    def __init__(self, serializer: Union[None, Serializers]):
        self.dump = None
        self.dumps = None
        self.load = None
        self.loads = None
        self._serializer = None
        self.serializer = serializer

    @property
    def serializer(self):
        return self._serializer

    @serializer.setter
    def serializer(self, serializer: Serializers):
        self._serializer = serializer
        if self._serializer is None:
            self.dump = None
            self.dumps = None
            self.load = None
            self.loads = None
        else:
            serializer_name, serializer_tags = SERIALIZERS_DESCRIPTION[self._serializer]
            self.dump = getattr(self, '_{}__dump'.format(serializer_name))
            self.dumps = getattr(self, '_{}__dumps'.format(serializer_name))
            self.load = getattr(self, '_{}__load'.format(serializer_name))
            self.loads = getattr(self, '_{}__loads'.format(serializer_name))

    # json
    @staticmethod
    def _json__dump(*args, **kwargs):
        json.dump(*args, **kwargs)

    @staticmethod
    def _json__dumps(*args, **kwargs):
        return json.dumps(*args, **kwargs)

    @staticmethod
    def _json__load(*args, **kwargs):
        return json.load(*args, **kwargs)

    @staticmethod
    def _json__loads(*args, **kwargs):
        return json.loads(*args, **kwargs)

    # simplejson
    @staticmethod
    def _simplejson__dump(*args, **kwargs):
        simplejson.dump(*args, **kwargs)

    @staticmethod
    def _simplejson__dumps(*args, **kwargs):
        return simplejson.dumps(*args, **kwargs)

    @staticmethod
    def _simplejson__load(*args, **kwargs):
        return simplejson.load(*args, **kwargs)

    @staticmethod
    def _simplejson__loads(*args, **kwargs):
        return simplejson.loads(*args, **kwargs)

    # ujson
    @staticmethod
    def _ujson__dump(*args, **kwargs):
        for unnecessary_parameter in ujson_dump_skip_parameters:
            kwargs.pop(unnecessary_parameter, None)
        ujson.dump(*args, **kwargs)

    @staticmethod
    def _ujson__dumps(*args, **kwargs):
        for unnecessary_parameter in ujson_dumps_skip_parameters:
            kwargs.pop(unnecessary_parameter, None)
        return ujson.dumps(*args, **kwargs)

    @staticmethod
    def _ujson__load(*args, **kwargs):
        for unnecessary_parameter in ujson_load_skip_parameters:
            kwargs.pop(unnecessary_parameter, None)
        return ujson.load(*args, **kwargs)

    @staticmethod
    def _ujson__loads(*args, **kwargs):
        for unnecessary_parameter in ujson_loads_skip_parameters:
            kwargs.pop(unnecessary_parameter, None)
        return ujson.loads(*args, **kwargs)

    # orjson
    @staticmethod
    def _orjson__dump(*args, **kwargs):
        for unnecessary_parameter in orjson_dump_skip_parameters:
            kwargs.pop(unnecessary_parameter, None)
        kwargs['option'] = orjson.OPT_NON_STR_KEYS
        orjson.dump(*args, **kwargs)

    @staticmethod
    def _orjson__dumps(*args, **kwargs):
        for unnecessary_parameter in orjson_dumps_skip_parameters:
            kwargs.pop(unnecessary_parameter, None)
        kwargs['option'] = orjson.OPT_NON_STR_KEYS
        return orjson.dumps(*args, **kwargs).decode('utf-8')

    @staticmethod
    def _orjson__load(*args, **kwargs):
        for unnecessary_parameter in orjson_load_skip_parameters:
            kwargs.pop(unnecessary_parameter, None)
        return orjson.load(*args, **kwargs)

    @staticmethod
    def _orjson__loads(*args, **kwargs):
        for unnecessary_parameter in orjson_loads_skip_parameters:
            kwargs.pop(unnecessary_parameter, None)
        return orjson.loads(*args, **kwargs)

    # tnetstring
    @staticmethod
    def _tnetstring__dump(*args, **kwargs):
        tnetstring.dump(*args, **kwargs)

    @staticmethod
    def _tnetstring__dumps(*args, **kwargs):
        return tnetstring.dumps(*args, **kwargs)

    @staticmethod
    def _tnetstring__load(*args, **kwargs):
        return tnetstring.load(*args, **kwargs)

    @staticmethod
    def _tnetstring__loads(*args, **kwargs):
        return tnetstring.loads(*args, **kwargs)

    # msgpack
    @staticmethod
    def _msgpack__dump(*args, **kwargs):
        result_kwargs = {
            # 'encoding': 'utf-8',  # PendingDeprecationWarning: encoding is deprecated.
            'use_bin_type': True,
        }
        result_kwargs.update(kwargs)
        msgpack.dump(*args, **result_kwargs)

    @staticmethod
    def _msgpack__dumps(*args, **kwargs):
        result_kwargs = {
            # 'encoding': 'utf-8',  # PendingDeprecationWarning: encoding is deprecated.
            'use_bin_type': True,
        }
        result_kwargs.update(kwargs)
        return msgpack.dumps(*args, **result_kwargs)

    @staticmethod
    def _msgpack__load(*args, **kwargs):
        result_kwargs = {
            # 'encoding': 'utf-8',  # PendingDeprecationWarning: encoding is deprecated, Use raw=False instead.
            'raw': False,
            'use_list': True,
            'strict_map_key': False,
        }
        result_kwargs.update(kwargs)
        with DisableGC():
            result = msgpack.load(*args, **result_kwargs)
        
        return result

    @staticmethod
    def _msgpack__loads(*args, **kwargs):
        result_kwargs = {
            # 'encoding': 'utf-8',  # PendingDeprecationWarning: encoding is deprecated, Use raw=False instead.
            'raw': False,
            'use_list': True,
            'strict_map_key': False,
        }
        result_kwargs.update(kwargs)
        with DisableGC():
            result = msgpack.loads(*args, **result_kwargs)
        
        return result

    # msgpack_fast
    @staticmethod
    def _msgpack_fast__dump(*args, **kwargs):
        result_kwargs = {
            'encoding': 'utf-8',
            'use_bin_type': True
        }
        result_kwargs.update(kwargs)
        msgpack.dump(*args, **result_kwargs)

    @staticmethod
    def _msgpack_fast__dumps(*args, **kwargs):
        result_kwargs = {
            'encoding': 'utf-8',
            'use_bin_type': True
        }
        result_kwargs.update(kwargs)
        return msgpack.dumps(*args, **result_kwargs)

    @staticmethod
    def _msgpack_fast__load(*args, **kwargs):
        result_kwargs = {
            'encoding': 'utf-8',
            'raw': False,
            'use_list': False
        }
        result_kwargs.update(kwargs)
        with DisableGC():
            result = msgpack.load(*args, **result_kwargs)
        
        return result

    @staticmethod
    def _msgpack_fast__loads(*args, **kwargs):
        result_kwargs = {
            'encoding': 'utf-8',
            'raw': False,
            'use_list': False
        }
        result_kwargs.update(kwargs)
        with DisableGC():
            result = msgpack.loads(*args, **result_kwargs)
        
        return result

    # msgspec_messagepack
    @staticmethod
    def _msgspec_messagepack__dump(obj, fp, *args, **kwargs):
        fp.write(msgspec.msgpack.encode(obj, *args, **kwargs))

    @staticmethod
    def _msgspec_messagepack__dumps(*args, **kwargs):
        return msgspec.msgpack.encode(*args, **kwargs)

    @staticmethod
    def _msgspec_messagepack__load(fp, *args, **kwargs):
        with DisableGC():
            result = msgspec.msgpack.decode(fp.read(), *args, **kwargs)
        
        return result

    @staticmethod
    def _msgspec_messagepack__loads(*args, **kwargs):
        with DisableGC():
            result = msgspec.msgpack.decode(*args, **kwargs)
        
        return result

    # msgspec_json
    @staticmethod
    def _msgspec_json__dump(obj, fp, *args, **kwargs):
        fp.write(msgspec.json.encode(obj, *args, **kwargs))

    @staticmethod
    def _msgspec_json__dumps(*args, **kwargs):
        return msgspec.json.encode(*args, **kwargs)

    @staticmethod
    def _msgspec_json__load(fp, *args, **kwargs):
        with DisableGC():
            result = msgspec.json.decode(fp.read(), *args, **kwargs)
        
        return result

    @staticmethod
    def _msgspec_json__loads(*args, **kwargs):
        with DisableGC():
            result = msgspec.json.decode(*args, **kwargs)
        
        return result

    # msgspec_yaml
    @staticmethod
    def _msgspec_yaml__dump(obj, fp, *args, **kwargs):
        fp.write(msgspec.yaml.encode(obj, *args, **kwargs))

    @staticmethod
    def _msgspec_yaml__dumps(*args, **kwargs):
        return msgspec.yaml.encode(*args, **kwargs)

    @staticmethod
    def _msgspec_yaml__load(fp, *args, **kwargs):
        with DisableGC():
            result = msgspec.yaml.decode(fp.read(), *args, **kwargs)
        
        return result

    @staticmethod
    def _msgspec_yaml__loads(*args, **kwargs):
        with DisableGC():
            result = msgspec.yaml.decode(*args, **kwargs)
        
        return result

    # msgspec_toml
    @staticmethod
    def _msgspec_toml__dump(obj, fp, *args, **kwargs):
        fp.write(msgspec.toml.encode(obj, *args, **kwargs))

    @staticmethod
    def _msgspec_toml__dumps(*args, **kwargs):
        return msgspec.toml.encode(*args, **kwargs)

    @staticmethod
    def _msgspec_toml__load(fp, *args, **kwargs):
        with DisableGC():
            result = msgspec.toml.decode(fp.read(), *args, **kwargs)
        
        return result

    @staticmethod
    def _msgspec_toml__loads(*args, **kwargs):
        with DisableGC():
            result = msgspec.toml.decode(*args, **kwargs)
        
        return result

    # cbor
    @staticmethod
    def _cbor__dump(*args, **kwargs):
        cbor.dump(*args, **kwargs)

    @staticmethod
    def _cbor__dumps(*args, **kwargs):
        return cbor.dumps(*args, **kwargs)

    @staticmethod
    def _cbor__load(*args, **kwargs):
        cbor.load(*args, **kwargs)

    @staticmethod
    def _cbor__loads(*args, **kwargs):
        return cbor.loads(*args, **kwargs)

    # cbor2
    @staticmethod
    def _cbor2__dump(*args, **kwargs):
        cbor2.dump(*args, **kwargs)

    @staticmethod
    def _cbor2__dumps(*args, **kwargs):
        return cbor2.dumps(*args, **kwargs)

    @staticmethod
    def _cbor2__load(*args, **kwargs):
        cbor2.load(*args, **kwargs)

    @staticmethod
    def _cbor2__loads(*args, **kwargs):
        return cbor2.loads(*args, **kwargs)

    # marshal
    @staticmethod
    def _marshal__dump(*args, **kwargs):
        if len(args) == 2:
            result_args = args
        else:
            result_args = (args[0], args[1], 4)
        marshal.dump(*result_args, **kwargs)

    @staticmethod
    def _marshal__dumps(*args, **kwargs):
        if len(args) == 1:
            result_args = args
        else:
            result_args = (args[0], 4)
        return marshal.dumps(*result_args, **kwargs)

    @staticmethod
    def _marshal__load(*args, **kwargs):
        return marshal.load(*args, **kwargs)

    @staticmethod
    def _marshal__loads(*args, **kwargs):
        return marshal.loads(*args, **kwargs)

    # pickle
    @staticmethod
    def _pickle__dump(*args, **kwargs):
        result_kwargs = {'protocol': -1}
        result_kwargs.update(kwargs)
        pickle.dump(*args, **result_kwargs)

    @staticmethod
    def _pickle__dumps(*args, **kwargs):
        result_kwargs = {'protocol': -1}
        result_kwargs.update(kwargs)
        return pickle.dumps(*args, **result_kwargs)

    @staticmethod
    def _pickle__load(*args, **kwargs):
        return pickle.load(*args, **kwargs)

    @staticmethod
    def _pickle__loads(*args, **kwargs):
        return pickle.loads(*args, **kwargs)

    # cPickle
    @staticmethod
    def _cpickle__dump(*args, **kwargs):
        result_kwargs = {'protocol': -1}
        result_kwargs.update(kwargs)
        cPickle.dump(*args, **result_kwargs)

    @staticmethod
    def _cpickle__dumps(*args, **kwargs):
        result_kwargs = {'protocol': -1}
        result_kwargs.update(kwargs)
        return cPickle.dumps(*args, **result_kwargs)

    @staticmethod
    def _cpickle__load(*args, **kwargs):
        return cPickle.load(*args, **kwargs)

    @staticmethod
    def _cpickle__loads(*args, **kwargs):
        return cPickle.loads(*args, **kwargs)

    # cloudpickle
    @staticmethod
    def _cloudpickle__dump(*args, **kwargs):
        result_kwargs = {'protocol': -1}
        result_kwargs.update(kwargs)
        cloudpickle.dump(*args, **result_kwargs)

    @staticmethod
    def _cloudpickle__dumps(*args, **kwargs):
        result_kwargs = {'protocol': -1}
        result_kwargs.update(kwargs)
        return cloudpickle.dumps(*args, **result_kwargs)

    @staticmethod
    def _cloudpickle__load(*args, **kwargs):
        return cloudpickle.load(*args, **kwargs)

    @staticmethod
    def _cloudpickle__loads(*args, **kwargs):
        return cloudpickle.loads(*args, **kwargs)


def get_an_appropriate_serializers(desired_features: SerializerFeatures) -> Set[Serializers]:
    appropriate_serializers = set()
    for serializer_type in EXISTING_SERIALIZERS:
        serializer_features = SERIALIZERS_DESCRIPTION[serializer_type][1]
        if desired_features.issubset(serializer_features):
            appropriate_serializers.add(serializer_type)
    return appropriate_serializers


SerializerPerformance = float
SerializerFootprint = int


def serializer_benchmark(serializer_type: Serializers, test_data: Optional[Any], test_time: float = 1.0
                         ) -> Tuple[SerializerPerformance, SerializerFootprint]:
    serializer = Serializer(serializer_type)
    measurements: List[float] = list()
    tr = Tracer(test_time)
    while tr.iter():
        start_time = perf_counter()
        serializer.loads(serializer.dumps(test_data))
        measurements.append(perf_counter() - start_time)

    best_measurement: float = min(measurements)
    best_performance: Optional[float] = (1 / best_measurement) if best_measurement else None
    data_dump = serializer.dumps(test_data)
    return tr.iter_per_time_unit, best_performance, len(data_dump)


def get_most_efficient_serializers(desired_features: SerializerFeatures,
                                   test_data: Optional[Any],
                                   test_time: float = 1.0
                                   ) -> Tuple[Set[Serializers],
                                              Set[Tuple[Serializers, SerializerPerformance, SerializerFootprint]]]:
    # TODO: сделать класс, который бы подбирал количество итераций под нужное время выдержки в секундах (float)
    # это необходимо для того чтобы правильно протестировать производительность под PyPy
    # в дальнейшем этот функционал стоит перенести в модуль performance_test_lib
    # TODO: make some caching algorithm. Lru can not be used since it requires all function parameters be hashable and both desired_features and test_data are not

    appropriate_serializers = get_an_appropriate_serializers(desired_features)

    benchmark_results = dict()  # type: Dict[float, Set[Serializers]]
    serializers_data = set()  # type: Set[Tuple[Serializers, SerializerPerformance, SerializerFootprint]]
    for serializer_type in appropriate_serializers:
        with DisableGC():
            performance, best_performance, dump_size = serializer_benchmark(serializer_type, test_data, test_time)
        
        if best_performance not in benchmark_results:
            benchmark_results[best_performance] = set()
        
        benchmark_results[best_performance].add(serializer_type)
        serializers_data.add((serializer_type, best_performance, dump_size))
    
    best_performance = max(benchmark_results)
    best_serializers = benchmark_results[best_performance]
    return best_serializers, serializers_data


def best_serializer(desired_features: SerializerFeatures,
                                   test_data: Optional[Any],
                                   test_time: float = 1.0
                                   ) -> Serializer:
    result = get_most_efficient_serializers(
        desired_features,
        test_data,
        test_time)
    best_serializers, _ = result
    return Serializer(best_serializers.pop())


class TestDataType(enum.Enum):
    small = 0
    deep_small = 1
    large = 2
    deep_large = 3


def test_data_factory(test_data_type: TestDataType):
    if TestDataType.small == test_data_type:
        return {
            1: 'Hello',
            2: ('W', 0),
            3: [
                'r',
                1,
                {
                    'd': '.',
                    'd'*2: {
                        43: [0]*2,
                        15: {
                            'world': 42
                        }
                    }
                }
            ]*2,
            'To all!': '!!1'
        }
    elif TestDataType.deep_small == test_data_type:
        return {
            1: 'Hello',
            2: ('W', 0),
            3: [
                'r',
                1,
                {
                    'd': '.',
                    'd'*2: {
                        43: [0]*1000,
                        15: {
                            'world': 42
                        }
                    }
                }
            ]*20,
            'To all!': '!!1'
        }
    elif TestDataType.large == test_data_type:
        return {
            1: 'Hello'*100,
            2: ('W', 0),
            3: [
                'r',
                1,
                {
                    'd': '.',
                    'd'*100: {
                        43: [0]*2,
                        15: {
                            'world'*100: 42
                        }
                    }
                }
            ]*2,
            'To all!': '!!1'*100
        }
    elif TestDataType.deep_large == test_data_type:
        return {
            1: 'Hello'*100,
            2: ('W', 0),
            3: [
                'r',
                1,
                {
                    'd': '.',
                    'd'*100: {
                        43: [0]*1000,
                        15: {
                            'world'*100: 42
                        }
                    }
                }
            ]*20,
            'To all!': '!!1'*100
        }


@lru_cache(maxsize=100)
def best_serializer_for_standard_data(desired_features_tuple: SerializerFeaturesTuple,
                                   test_data_type: Optional[TestDataType] = None,
                                   test_time: float = 1.0
                                   ) -> Serializer:
    test_data_type = TestDataType.small if test_data_type is None else test_data_type
    return best_serializer(set(desired_features_tuple), test_data_factory(test_data_type), test_time)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from inspect import isclass
from cengal.introspection.inspect import get_exception
from cengal.code_flow_control.args_manager import EntityArgsHolder
from typing import Any, Callable, NoReturn, Set, Dict, Type, TypeVar, Generic, Optional, Union, Sequence
# from enum import Enum


class EssenceModelException(Exception):
    pass


class IncompatibleEssenceModelError(RuntimeError, EssenceModelException):
    # Must be raised by constructor of EssenceInterface class if incompatible essence_model was given
    pass


class UnsuitableEssenceInterfaceError(RuntimeError, EssenceModelException):
    pass


class EssenceInterfaceIsNotApplicableError(UnsuitableEssenceInterfaceError, EssenceModelException):
    # Must be raised by EssenceModel.__call__ when requested EssenceInterface is not active (applicable)
    pass


class EssenceInterfaceIsNotRegisteredError(UnsuitableEssenceInterfaceError, EssenceModelException):
    # Must be raised by EssenceModel.__call__ when requested EssenceInterface is not registered
    pass


class EssenceModelCanNotInjectSelfError(RuntimeError, EssenceModelException):
    """Must be raised by EssenceModel.emi_inject_model when an object of a class A(EssenceModel) tries
    to inject class A - type(self)

    This behavior is restricted since it may lead to: 
        - endless recursion
        - interface collisions when high order model will include several instances of the same interface class"""
    pass


class EssenceModelCanNotBeInjectedError(RuntimeError, EssenceModelException):
    # Must be raised by EssenceModel.emi_inject_model if injectable model can not be injected (incompatible)
    pass


class EssenceModelIsNotInjectedError(RuntimeError, EssenceModelException):
    # Must be raised by EssenceModel.emi_injected_model if not injected model was requested
    pass


class IncompatibleHighOrderEssenceModelError(RuntimeError, EssenceModelException):
    # Must be raised by EssenceModel.emu_behave_as_unknown_model method if incompatible on attempt to inject it to
    # incompatible high order model
    pass


class UnknownEssenceModeBehaviorWasNotImplementedProperlyError(NotImplementedError, EssenceModelException):
    pass


class EssenceModelInheritanceAbstract:
    def __call__(self, interface_class: Type['EssenceInterface'], worker: Optional[Callable] = None, failed_worker: Optional[Callable] = None, *args, **kwargs) -> Any:
        raise NotImplementedError

    def em_interface(self, interface_class: Type['EssenceInterface']) -> 'EssenceInterface':
        raise NotImplementedError

    def em_has_interface(self, interface_class: Type['EssenceInterface']) -> bool:
        raise NotImplementedError

    def em_interface_active(self, interface_class: Type['EssenceInterface']) -> bool:
        raise NotImplementedError

    def em_active_interfaces(self) -> Set[Type['EssenceInterface']]:
        raise NotImplementedError

    def em_all_interfaces(self) -> Set[Type['EssenceInterface']]:
        raise NotImplementedError

    def _em_check_applicability_of_interfaces(self) -> None:
        raise NotImplementedError

    def em_on_model_updated(self, interface_class: Type['EssenceInterface'], *args, **kwargs):
        raise NotImplementedError

    def em_add_interface(self, interface_class: Type['EssenceInterface'], *args, **kwargs) -> bool:
        raise NotImplementedError

    def em_remove_interface(self, interface_class: Type['EssenceInterface']) -> bool:
        raise NotImplementedError


class EssenceModelInjectionAbstract:
    def emi_on_registered_in_high_order_model(self, high_order_model: 'EssenceModelInjectionAbstract'):
        raise NotImplementedError

    def emi_on_unregistering_from_high_order_model(self):
        raise NotImplementedError

    def _emi_notify_high_order_model_about_self_update(self, *args, **kwargs):
        raise NotImplementedError

    def emi_inject_model(self, essence_model: 'EssenceModelInjectionAbstract') -> bool:
        raise NotImplementedError

    def emi_injected_models(self):
        raise NotImplementedError

    def emi_injected_model(self, essence_model_class: Type['EssenceModelInjectionAbstract']):
        raise NotImplementedError

    def emi_on_injected_model_updated(
            self, essence_model_class: Type['EssenceModelInjectionAbstract'], *args, **kwargs):
        raise NotImplementedError

    def emi_remove_injected_model(self, essence_model_class: Type['EssenceModelInjectionAbstract']):
        raise NotImplementedError


class EssenceModelUnknownInjectionAbstract:
    def emu_is_compatible_high_order_model(self, high_order_model_class: Type['EssenceModelInjectionAbstract']) -> bool:
        raise NotImplementedError
    
    def emu_behave_as_unknown_model(self):
        raise NotImplementedError

    def emu_on_behave_as_unknown_model(self):
        raise NotImplementedError

    def _emu_register_on_model_changed_callback(self, requester: 'EssenceModelUnknownInjectionAbstract'):
        raise NotImplementedError

    def _emu_deregister_on_model_changed_callback(self, requester: 'EssenceModelUnknownInjectionAbstract'):
        raise NotImplementedError
    
    def _emu_notify_unknown_models_about_self_update(self):
        raise NotImplementedError

    def emu_on_model_changed_callback(self):
        raise NotImplementedError

    def emu_is_in_unknown_model_behavior(self):
        raise NotImplementedError


class EssenceModel(EssenceModelInheritanceAbstract, EssenceModelInjectionAbstract, EssenceModelUnknownInjectionAbstract):
    """Must contain related data in a consistent state.
    
    In order to do this, you must reload `em_on_model_updated()` and `emi_on_injected_model_updated()` methods.
    Your interfaces can provide some appropriate information though an additional `parameters of em_on_model_updated()`
    and `emi_on_injected_model_updated()` methods.
    
    Do not forget to:
      - run `self._emi_notify_high_order_model_about_self_update(type(self), interface_class, *args, **kwargs)` at
          the end of your `em_on_model_updated()`
      - run `self._emi_notify_high_order_model_about_self_update(type(self), essence_model_class, *args, **kwargs)` at
          the end of your `emi_on_injected_model_updated()`"""

    emi_compatible_injectable_essence_model_classes: Set[Type['EssenceModel']] = set()
    # emu_compatible_high_order_essence_model_class: Optional[Type['EssenceModel']] = None
    emu_compatible_high_order_essence_model_classes: Set[Type['EssenceModel']] = set()

    def __init__(self):
        self.__em_interfaces: Dict[Type['EssenceInterface'], 'EssenceInterface'] = dict()
        self.__em_possible_interfaces: Dict[Type['EssenceInterface'], 'EssenceInterface'] = dict()
        self.__emi_injected_models: Dict[Type['EssenceModel'], 'EssenceModel'] = dict()
        self.__emi_raise_on_uninjectable_model: bool = True
        self.__emi_high_order_model: Optional['EssenceModel'] = None
        self.__emu_in_unknown_model_behavior: bool = False
        self.__emu_raise_on_incompatible_high_order_model: bool = True
        self.__emu_unknown_injected_models: Dict[Type['EssenceModel'], 'EssenceModel'] = dict()

    def em_interface(self, interface_class: Type['EssenceInterface']) -> 'EssenceInterface':
        """Should be called in order to get needed model interface"""
        if interface_class in self.__em_interfaces:
            return self.__em_interfaces[interface_class]
        else:
            injected_model_interface = None
            for injected_model_class, injected_model in self.__emi_injected_models.items():
                model: 'EssenceModel' = injected_model
                if model.em_interface_active(interface_class):
                    injected_model_interface = model.em_interface(interface_class)
                    break
            if injected_model_interface:
                return injected_model_interface
            elif interface_class in self.__em_possible_interfaces:
                raise EssenceInterfaceIsNotApplicableError
            else:
                raise EssenceInterfaceIsNotRegisteredError

    def __call__(self, interface_class: Type['EssenceInterface'], worker: Callable, failed_worker: Optional[Callable] = None, *args, **kwargs) -> Any:
        interface = None
        exception = None
        try:
            interface = self.em_interface(interface_class)
        except (EssenceInterfaceIsNotApplicableError, EssenceInterfaceIsNotRegisteredError) as exc:
            exception = get_exception()
        
        if interface:
            return worker(interface, *args, **kwargs)
        else:
            if failed_worker:
                return failed_worker(interface, exception, *args, **kwargs)
            else:
                return None
    
    def em_has_interface(self, interface_class: Type['EssenceInterface']) -> bool:
        has_own_interface = (interface_class in self.__em_interfaces) or \
                            (interface_class in self.__em_possible_interfaces)
        one_of_injected_models_has_interface = False
        for injected_model_class, injected_model in self.__emi_injected_models.items():
            model: 'EssenceModel' = injected_model
            if model.em_has_interface(interface_class) or model.em_interface_active(interface_class):
                one_of_injected_models_has_interface = True
                break
        return has_own_interface or one_of_injected_models_has_interface

    def em_interface_active(self, interface_class: Type['EssenceInterface']) -> bool:
        own_interface_active = interface_class in self.__em_interfaces
        one_of_injected_models_has_active_interface = False
        for injected_model_class, injected_model in self.__emi_injected_models.items():
            model: 'EssenceModel' = injected_model
            if model.em_interface_active(interface_class):
                one_of_injected_models_has_active_interface = True
                break
        return own_interface_active or one_of_injected_models_has_active_interface

    def em_active_interfaces(self) -> Set[Type['EssenceInterface']]:
        own_active_interfaces = set(self.__em_interfaces)
        injected_models_active_interfaces: Set[Type['EssenceInterface']] = set()
        for injected_model_class, injected_model in self.__emi_injected_models.items():
            model: 'EssenceModel' = injected_model
            injected_models_active_interfaces.update(model.em_active_interfaces())
        return own_active_interfaces | injected_models_active_interfaces

    def em_all_interfaces(self) -> Set[Type['EssenceInterface']]:
        own_all_interfaces = set(self.__em_interfaces) | set(self.__em_possible_interfaces)
        injected_models_all_interfaces: Set[Type['EssenceInterface']] = set()
        for injected_model_class, injected_model in self.__emi_injected_models.items():
            model: 'EssenceModel' = injected_model
            injected_models_all_interfaces.update(model.em_all_interfaces())
        return own_all_interfaces | injected_models_all_interfaces

    def _em_check_applicability_of_interfaces(self) -> None:
        new_interfaces = dict()
        new_possible_interfaces = dict()

        for interface_class, interface in self.__em_interfaces.items():
            if interface._applicable_impl():
                new_interfaces[interface_class] = interface
            else:
                new_possible_interfaces[interface_class] = interface

        for interface_class, interface in self.__em_possible_interfaces.items():
            if interface._applicable_impl():
                new_interfaces[interface_class] = interface
            else:
                new_possible_interfaces[interface_class] = interface

        self.__em_interfaces = new_interfaces
        self.__em_possible_interfaces = new_possible_interfaces

    def em_on_model_updated(self, interface_class: Type['EssenceInterface'], *args, **kwargs):
        """Must be run by EssenceInterface (by running EssenceInterface.notify_model_about_change method) after changing
          model's data. It is enough to run in once per a method - at the end of the method work.
        In 'super' in method of inherit class should be run at the end of the method"""
        self._em_check_applicability_of_interfaces()
        self._emu_notify_unknown_models_about_self_update()
        self._emi_notify_high_order_model_about_self_update(type(self), interface_class, *args, **kwargs)

    def em_add_interface(self, interface_class: Type['EssenceInterface'], *args, **kwargs) -> bool:
        if (interface_class in self.__em_interfaces) or (interface_class in self.__em_possible_interfaces):
            return False
        else:
            interface = interface_class(self, *args, **kwargs)
            if interface._applicable_impl():
                self.__em_interfaces[interface_class] = interface
            else:
                self.__em_possible_interfaces[interface_class] = interface
            return True

    def em_remove_interface(self, interface_class: Type['EssenceInterface']) -> bool:
        if interface_class in self.__em_interfaces:
            del self.__em_interfaces[interface_class]
            return True
        elif interface_class in self.__em_possible_interfaces:
            del self.__em_possible_interfaces[interface_class]
            return True
        else:
            return False

    def emi_on_registered_in_high_order_model(self, high_order_model: 'EssenceModel'):
        """Will be called after high order model successfully registered this mode"""
        self.__emi_high_order_model = high_order_model

    def emi_on_unregistering_from_high_order_model(self):
        # Will be called before high order model actually unregistered this mode
        self.__emi_high_order_model = None
        self.__emu_in_unknown_model_behavior = False

    def emi_inject_model(self, essence_model: 'EssenceModel'):
        # Should use 'EssenceModel' instead of Type['EssenceModel'] since we should use result of fully constructed
        #   essence_model with all needed interfaces - result of an appropriate factory work
        essence_model_class = type(essence_model)
        if isinstance(essence_model, type(self)) or isinstance(self, essence_model_class):
            raise EssenceModelCanNotInjectSelfError
        
        injectable_model = False
        if essence_model_class in self.emi_compatible_injectable_essence_model_classes:
            injectable_model = True
        else:
            if essence_model.emu_is_compatible_high_order_model(type(self)):
                injectable_model = True
            else:
                if self.__emu_raise_on_incompatible_high_order_model:
                    raise IncompatibleHighOrderEssenceModelError
        
        injected = False
        if injectable_model:
            self.__emi_injected_models[essence_model_class] = essence_model
            essence_model.emi_on_registered_in_high_order_model(self)
            if essence_model_class not in self.emi_compatible_injectable_essence_model_classes:
                essence_model.emu_behave_as_unknown_model()
            
            injected = True
        else:
            if self.__emi_raise_on_uninjectable_model:
                raise EssenceModelCanNotBeInjectedError
            
        return injected

    def emi_injected_models(self):
        return set(self.__emi_injected_models)

    def emi_injected_model(self, essence_model_class: Type['EssenceModel']):
        if essence_model_class in self.__emi_injected_models:
            return self.__emi_injected_models[essence_model_class]
        else:
            raise EssenceModelIsNotInjectedError

    def emi_on_injected_model_updated(self, essence_model_class: Type['EssenceModel'], *args, **kwargs):
        # In 'super' in method of inherit class should be run at the end of the method
        # With deep injection it will be like (model_3, model_2, model_1, interface_1, arg_1, arg_2, arg_3)
        #   where `interface_1` is an interface of the `model_1`
        self._em_check_applicability_of_interfaces()
        self._emu_notify_unknown_models_about_self_update()
        self._emi_notify_high_order_model_about_self_update(type(self), essence_model_class, *args, **kwargs)

    def _emi_notify_high_order_model_about_self_update(self, *args, **kwargs):
        if self.__emi_high_order_model and (not self.__emu_in_unknown_model_behavior):
            high_order_model: 'EssenceModel' = self.__emi_high_order_model
            high_order_model.emi_on_injected_model_updated(*args, **kwargs)

    def emi_remove_injected_model(self, essence_model_class: Type['EssenceModel']):
        injected_model: 'EssenceModel' = self.__emi_injected_models[essence_model_class]
        self._emu_deregister_on_model_changed_callback(injected_model)
        injected_model.emi_on_unregistering_from_high_order_model()
        del self.__emi_injected_models[essence_model_class]

    def emu_is_compatible_high_order_model(self, high_order_model_class: Type['EssenceModelInjectionAbstract']) -> bool:
        """Might be overloaded in order to make this model compatible with more than one high order models

        Args:
            high_order_model_class (Type[): [description]

        Returns:
            bool: [description]
        """
        # return high_order_model_class == self.emu_compatible_high_order_essence_model_class
        return high_order_model_class in self.emu_compatible_high_order_essence_model_classes

    def emu_behave_as_unknown_model(self):
        # Should be called by a high order model for an unknown injected model
        # if type(self.__emi_high_order_model) != self.emu_compatible_high_order_essence_model_class:
        if not self.emu_is_compatible_high_order_model(self.__emi_high_order_model):
            raise IncompatibleHighOrderEssenceModelError
        self.__emi_high_order_model._emu_register_on_model_changed_callback(type(self))
        self.__emu_in_unknown_model_behavior = True
        self.emu_on_behave_as_unknown_model()

    def emu_on_behave_as_unknown_model(self):
        pass

    def _emu_register_on_model_changed_callback(self, requester: 'EssenceModelUnknownInjectionAbstract'):
        self.__emu_unknown_injected_models[type[requester]] = requester

    def _emu_deregister_on_model_changed_callback(self, requester: 'EssenceModelUnknownInjectionAbstract'):
        model_type = type(requester)
        if model_type in self.__emu_unknown_injected_models:
            del self.__emu_unknown_injected_models[model_type]

    def _emu_notify_unknown_models_about_self_update(self, *args, **kwargs):
        for umodel in self.__emu_unknown_injected_models.values():
            unknown_model: 'EssenceModel' = umodel
            unknown_model.emu_on_model_changed_callback()

    def emu_on_model_changed_callback(self):
        raise UnknownEssenceModeBehaviorWasNotImplementedProperlyError

    def emu_is_in_unknown_model_behavior(self):
        return self.__emu_in_unknown_model_behavior


Model = TypeVar('Model', bound=EssenceModel)


class EssenceInterface(Generic[Model]):
    essence_model_class: Type[Model] = EssenceModel

    def __init_subclass__(cls, /, essence_model_class: Optional[Type[Model]] = None, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.essence_model_class = EssenceModel if essence_model_class is None else essence_model_class

    def __init__(self, essence_model: Model, *args, **kwargs):
        self.essence_model: Model = essence_model
        self.em: Model = essence_model
        self._on_applicability_changed_handlers: Set[Callable] = set()
        self._applicability_state: bool = None
        self.__check_essence_mode_type()

    def __check_essence_mode_type(self):
        if not isinstance(self.essence_model, self.essence_model_class):
            raise IncompatibleEssenceModelError

    def applicable(self) -> bool:
        return True

    def _applicable_impl(self) -> bool:
        result: bool = self.applicable()
        if result != self._applicability_state:
            self._applicability_state = result
            for handler in self._on_applicability_changed_handlers:
                handler(self)
        
        return result
    
    def add_on_applicability_changed_handler(self, handler: Callable):
        self._on_applicability_changed_handlers.add(handler)
    
    def discard_on_applicability_changed_handler(self, handler: Callable):
        self._on_applicability_changed_handlers.discard(handler)

    def notify_model_about_change(self, *args, **kwargs):
        # Must be run by EssenceInterface's methods if and after changing model's data. It is enough to run in once per
        #   a method - at the end of the method work.
        self.essence_model.em_on_model_updated(type(self), *args, **kwargs)


def essence_model_changer(func):
    def wrapper(self: EssenceInterface, *args, **kwargs):
        try:
            func(self, *args, **kwargs)
        finally:
            self.notify_model_about_change(*args, **kwargs)

    return wrapper


em_changer = essence_model_changer


class EssenceModelFactoryExample:
    def __call__(self, *args, **kwargs):
        model: EssenceModel = EssenceModel()
        model.em_add_interface(EssenceInterface)
        return model


def simple_essence_model_factory(
        model: Union[Type[EssenceModel], EntityArgsHolder], 
        interfaces: Union[
                Type[EssenceInterface], 
                EntityArgsHolder, 
                Sequence[Union[Type[EssenceInterface], EntityArgsHolder]]
            ]
    ) -> EssenceModel:
    model_instance: EssenceModel = model()
    if isinstance(interfaces, EntityArgsHolder) or (isinstance(interfaces, type) and issubclass(interfaces, EssenceInterface)):
        interfaces = (interfaces,)
    
    for interface in interfaces:
        if isinstance(interface, type):
            if issubclass(interface, EssenceInterface):
                model_instance.em_add_interface(interface)
        elif isinstance(interface, EntityArgsHolder):
            model_instance.em_add_interface(*interface.entity_args_kwargs())
        else:
            raise TypeError(f'Wrong interface type: {type(interface)}')

    return model_instance


simple_em_factory = simple_essence_model_factory



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cengal.data_manipulation.tree_traversal import *
from cengal.data_containers.stack import *

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def objects_counter(data: AnAppropriateContainers):
    obj_sum = 0
    stack = Stack(TreeStackItem(data_2_tuple(data)))
    # stack = StackUni(TreeStackItem(data_2_tuple(data)), remove_on_pop=False)
    while stack:
        # print(obj_summ)
        # print(stack.top().node)
        # print(stack.top().child)
        # print()
        if stack.top().child is None:
            obj_sum += 1
            if isinstance(stack.top().node, (tuple, list, deque)):
                if stack.top().node:
                    stack.top().child = -1
                    continue
                else:
                    stack.pop()
                    continue
            else:
                stack.pop()
                continue
        else:
            stack.top().child += 1
            if len(stack.top().node) > stack.top().child:
                stack.push(TreeStackItem(data_2_tuple(stack.top().node[stack.top().child])))
                continue
            else:
                stack.pop()
                continue
    return obj_sum


def _objects_counter_recursive_impl(data: AnAppropriateContainers) -> int:
    obj_sum = 1
    if isinstance(data, (tuple, list, deque)):
        for item in data:
            obj_sum += _objects_counter_recursive_impl(data_2_tuple(item))
    return obj_sum


def objects_counter_recursive(data: AnAppropriateContainers) -> int:
    return _objects_counter_recursive_impl(data_2_tuple(data))


def object_counter_uni(data: AnAppropriateContainers, on_changed_to_drop_in_replacement: Optional[Callable] = None) -> int:
    return recursion_insureness(objects_counter_recursive, objects_counter, on_changed_to_drop_in_replacement, data)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from typing import Dict, Tuple, Union, Iterable, List, Callable
from cengal.system import CENGAL_IS_IN_BUILD_MODE
from cengal.build_tools.find_and_prepare_cython_modules import find_and_prepare_cython_modules as find_and_prepare_cython_modules_orig, \
    sdist as sdist_orig, build as build_orig, build_ext as build_ext_orig, \
    BuildConfig as BuildConfig_orig, Distribution, \
    find_good_packages as find_good_packages_orig, find_package_data as find_package_data_orig
from cengal.file_system.path_manager import path_relative_to_src, RelativePath, get_relative_path_part, sep


PACKAGE_SRC_RELATIVE_PATH: str = './cengal'
MODULES_TO_IGNORE = {'cengal._template_module'}
PYTHON_2_MODULES = {'cengal.cross_version.console_print.python2'}
EXCLUDE_FROM_PACKAGES = [
            'tests', 'tests.*', 'multiproject', 'multiproject.*', 'examples', 'examples.*', 'docs', 'docs.*', 'benchmarks', 'benchmarks.*', 
            'cengal.egg-info', 'cengal.egg-info.*', 'dist', '.eggs', '.idea', '.vscode', 'package',
        ]


def repo_root_rel() -> RelativePath:
    return RelativePath(path_relative_to_src('../..'))


def cengal_root_rel() -> RelativePath:
    return RelativePath(repo_root_rel()(PACKAGE_SRC_RELATIVE_PATH))


def find_good_packages():
    return find_good_packages_orig(where=repo_root_rel()(''), exclude=EXCLUDE_FROM_PACKAGES, modules_to_ignore=MODULES_TO_IGNORE, python_2_modules=PYTHON_2_MODULES)


def find_package_data() -> Tuple[Dict[str, List[str]], List[str]]:
    return find_package_data_orig(PACKAGE_SRC_RELATIVE_PATH, find_good_packages(), root_rel=repo_root_rel())


def find_and_prepare_cython_modules(depth = 1):
    depth = depth or 0
    depth += 1
    additional_cflags: Dict[str, Tuple[bool, Union[str, int]]] = {
        'CENGAL_IS_IN_BUILD_MODE': CENGAL_IS_IN_BUILD_MODE,
    }
    return find_and_prepare_cython_modules_orig(PACKAGE_SRC_RELATIVE_PATH, additional_cflags, depth, root_rel=repo_root_rel())


class BuildConfig(BuildConfig_orig):
    def __init__(self) -> None:
        super().__init__()
        self.package_build_is_in_debug_mode: str = 'CENGAL_BUILD_IS_IN_DEBUG_MODE'
        self.package_build_is_in_sdist_mode: str = 'CENGAL_BUILD_IS_IN_SDIST_MODE'
        self.additional_pyx_flags: Dict[str, Tuple[bool, Union[str, int]]] = {
            'CENGAL_IS_IN_BUILD_MODE': CENGAL_IS_IN_BUILD_MODE,
        }
        self.find_package_data: Callable = find_package_data
        self.root_rel: RelativePath = repo_root_rel()


class sdist(sdist_orig):
    def __init__(self, dist: Distribution) -> None:
        super().__init__(dist, BuildConfig())


class build(build_orig):
    def __init__(self, dist: Distribution) -> None:
        super().__init__(dist, BuildConfig())


class build_ext(build_ext_orig):
    def __init__(self, dist: Distribution) -> None:
        super().__init__(dist, BuildConfig())



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.parallel_execution.asyncio.init_loop import init_loop, restore_default_loop
from cengal.parallel_execution.coroutines.integrations.uvloop import uvloop_install
from cengal.parallel_execution.asyncio.efficient_streams import *
from cengal.io.used_ports import *
from cengal.system import OS_TYPE

from typing import Dict
import asyncio
import json
from unittest import TestCase, main, skip, skipIf


# Raw protocol:

async def server_worker(server, reader, writer):
    reader: TcpStreamReader = reader
    writer: TcpStreamWriter = writer

    message_size_len_encoded = await reader.readexactly(8)
    message_size_len = int.from_bytes(message_size_len_encoded, 'little')
    data = await reader.readexactly(message_size_len)
    message_raw = bytes(data).decode()
    message_info: Dict = json.loads(message_raw)
    message: str = message_info['message']
    messages_num: int = message_info['messages_num']
    message_data: bytes = message.encode()

    writer.start_aw()
    try:
        while messages_num:
            messages_num -= 1
            writer.owrite(message_data)
            await writer.aw_drain_enough()
        
        await writer.full_drain()
    finally:
        writer.close()
        await writer.stop_aw()
        server.close()


async def server(port: int, server_ready: asyncio.Future):
    server = await TcpStreamManager().start_server(
        lambda reader, writer: server_worker(server, reader, writer), '127.0.0.1', port)

    async with server:
        server_ready.set_result(True)
        try:
            await server.serve_forever()
        except asyncio.exceptions.CancelledError:
            pass


async def client(port: int) -> bool:
    message: str = 'Hello World!'
    message_encoded: bytes = message.encode()
    messages_num: int = 1000
    message_info: Dict = {
        'message': message,
        'messages_num': messages_num
    }
    message_raw: str = json.dumps(message_info)
    data: bytes = message_raw.encode()
    message_size_len: int = len(data)
    message_size_len_encoded: bytes = message_size_len.to_bytes(8, 'little')

    reader, writer = await TcpStreamManager().open_connection(
        '127.0.0.1', port)

    writer.write(message_size_len_encoded)
    writer.write(data)
    await writer.drain()

    data_len_counter = 0
    data = None
    while data is None or data:
        data = await reader.read_nearly_max()
        data_len_counter += len(data)

    writer.close()
    return data_len_counter == (messages_num * len(message_encoded))


async def transmit(port: int) -> bool:
    server_ready: asyncio.Future = asyncio.Future()
    server_task = asyncio.create_task(server(port, server_ready))
    await server_ready
    client_task = asyncio.create_task(client(port))
    await asyncio.gather(server_task, client_task)
    result = client_task.result()
    return result


# Message protocol:

async def msg_server_worker(stream_manager: TcpStreamManager, server, reader, writer):
    if not await stream_manager.try_establish_message_protocol_server_side(reader, writer):
        raise RuntimeError('Failed to establish message protocol')

    reader: TcpStreamReader = reader
    writer: TcpStreamWriter = writer
    
    data: bytes = await reader.read_message()
    message_raw = bytes(data).decode()
    message_info: Dict = json.loads(message_raw)
    message: str = message_info['message']
    messages_num: int = message_info['messages_num']
    message_data: bytes = message.encode()

    with writer:
        async with writer.aw():
            while messages_num:
                messages_num -= 1
                writer.owrite_message(message_data)
                await writer.aw_drain_enough()
            
            await writer.full_drain()

    server.close()


async def msg_server(port: int, server_ready: asyncio.Future):
    stream_manager: TcpStreamManager = TcpStreamManager()
    server = await stream_manager.start_server(
        lambda reader, writer: msg_server_worker(stream_manager, server, reader, writer), '127.0.0.1', port)

    async with server:
        server_ready.set_result(True)
        try:
            await server.serve_forever()
        except asyncio.exceptions.CancelledError:
            pass


async def msg_client(port: int) -> bool:
    message: str = 'Hello World!'
    message_encoded: bytes = message.encode()
    messages_num: int = 1000
    message_info: Dict = {
        'message': message,
        'messages_num': messages_num
    }
    message_raw: str = json.dumps(message_info)
    data: bytes = message_raw.encode()

    stream_manager: TcpStreamManager = TcpStreamManager()
    reader, writer = await stream_manager.open_connection(
        '127.0.0.1', port)

    if not await stream_manager.try_establish_message_protocol_client_side(reader, writer):
        raise RuntimeError('Failed to establish message protocol')

    with writer:
        async with writer.aw():
            await writer.send_message(data)

            data_len_counter = 0
            data = None
            while data is None or data:
                try:
                    data = await reader.read_message()
                except asyncio.exceptions.IncompleteReadError:
                    data = bytes()
                
                data_len_counter += len(data)

    return data_len_counter == (messages_num * len(message_encoded))


async def msg_transmit(port: int) -> bool:
    server_ready: asyncio.Future = asyncio.Future()
    server_task = asyncio.create_task(msg_server(port, server_ready))
    await server_ready
    client_task = asyncio.create_task(msg_client(port))
    await asyncio.gather(server_task, client_task)
    result = client_task.result()
    return result


class TestTcpEfficientStream(TestCase):
    def setUp(self):
        restore_default_loop()
        init_loop()

    def test_transmit(self):
        port = purify_ports(used_ports().port(Protocol.tcp, used_ports().tables[Table.user], {PortStatus.no})())
        result = asyncio.run(transmit(port))
        self.assertTrue(result)

    def test_msg_transmit(self):
        port = purify_ports(used_ports().port(Protocol.tcp, used_ports().tables[Table.user], {PortStatus.no})())
        result = asyncio.run(msg_transmit(port))
        self.assertTrue(result)


def transmit_test():
    port = purify_ports(used_ports().port(Protocol.tcp, used_ports().tables[Table.user], {PortStatus.no})())
    result = asyncio.run(transmit(port))

def msg_transmit_test():
    port = purify_ports(used_ports().port(Protocol.tcp, used_ports().tables[Table.user], {PortStatus.no})())
    result = asyncio.run(msg_transmit(port))


def general_main():
    print(transmit_test())
    print(msg_transmit_test())


if __name__ == '__main__':
    main()
    # general_main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import socket
from contextlib import contextmanager
from cengal.base.classes import BaseClassSettings
from .base import *




__author__ = 'ButenkoMS <gtalk@butenkoms.space>'


class SimpleTcpLinkError(SimpleNetworkError):
    pass


class SimpleTcpLink:
    '''
    1 to 1 TCP connection.
    '''
    def __init__(self, settings: ConnectionSettings):
        '''
        Port should not be open to a external world!
        :param settings: ConnectionSettings()
        :return:
        '''
        self.settings = settings
        self.settings.check()
        self._block_state = True
        self._gate = None
        self._conn = None
        self._addr = None

        self.message_size_len = MESSAGE_SIZE_LEN
        self.server_answer__keyword_accepted = SERVER_ANSWER__KEYWORD_ACCEPTED
        self.use_nodelay_inet = False
        pass

    def connect(self):
        if ConnectionDirectionRole.server == self.settings.direction_role:
            self._gate = socket.socket(self.settings.socket_family, self.settings.socket_type,
                                       self.settings.socket_protocol, self.settings.socket_fileno)
            self._gate.bind(self.settings.socket_address)
            self._gate.listen(1)
            self._conn, self._addr = self._gate.accept()
            print('Connected by', self._addr)
            keyword_length_raw = self._conn.recv(self.message_size_len)
            keyword_length = int.from_bytes(keyword_length_raw, 'little')
            keyword = self._conn.recv(keyword_length)
            if keyword == self.settings.keyword:
                self._conn.sendall(self._pack_message(self.server_answer__keyword_accepted))
            else:
                # if we'll use loop, hang will be possible.
                raise SimpleTcpLinkError('Wrong keyword: {}. Should be: {}.'.format(
                    keyword, self.settings.keyword))
        elif ConnectionDirectionRole.client == self.settings.direction_role:
            self._conn = socket.socket(self.settings.socket_family, self.settings.socket_type,
                                       self.settings.socket_protocol, self.settings.socket_fileno)
            self._conn.connect(self.settings.socket_address)
            self.send_message_b(self.settings.keyword)
            if self.read_message_b() != self.server_answer__keyword_accepted:
                raise SimpleTcpLinkError('Server rejected the keyword: {}.'.format(
                    self.settings.keyword))
        else:
            raise NotImplementedError('Unknown ConnectionDirectionRole.')

        if self.use_nodelay_inet and (self._conn.family in INET_TYPE_CONNECTIONS):
            self._conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    def send_message_b(self, data):
        self._conn.sendall(len(data).to_bytes(self.message_size_len, 'little') + data)

    def read_message_b(self):
        return self._conn.recv(int.from_bytes(self._conn.recv(self.message_size_len), 'little'))

    def send_message(self, data):
        if self._block_state:
            self.send_message_b(data)
        else:
            pass
        pass

    def read_message(self):
        if self._block_state:
            return self.read_message_b()
        else:
            pass

    def message_io_iteration(self):
        pass

    def set_blocking(self, set_blocking_mode=True):
        pass

    def close(self):
        if self._conn is not None:
            self._conn.close()
        if self._gate is not None:
            self._gate.close()

    def _pack_message(self, data):
        return len(data).to_bytes(self.message_size_len, 'little') + data


@contextmanager
def simple_tcp_link_connect(simple_tcp_obj):
    try:
        simple_tcp_obj.connect()
        yield simple_tcp_obj
    except:
        raise
    finally:
        simple_tcp_obj.close()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Set, Dict, Tuple, Any, Hashable, Optional, FrozenSet, Callable


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


ItemID = Hashable
TagID = Hashable
UniqueItemID = Hashable


class UniqueIdGen:
    def __init__(self):
        self.index = 0

    def __call__(self):
        result = self.index
        self.index += 1
        return result


def tags_set_2_tuple(tags: Set[TagID]) -> Tuple[TagID]:
    tags_list = list(tags)
    tags_list.sort()
    return tuple(tags_list)


def default_growth_during_the_search_checker(): return True


def default_scheduler(): pass


class TagNotFound(Exception):
    pass


class DynamicTagTree:
    def __init__(self,
                 growth_during_the_search_is_allowed: Optional[Callable] = None,
                 scheduler: Optional[Callable] = None):
        self.growth_during_the_search_is_allowed: Callable = \
            growth_during_the_search_is_allowed or default_growth_during_the_search_checker
        self.scheduler: Callable = scheduler or default_scheduler
        self.uiidg = UniqueIdGen()
        self.item_by_uiid: Dict[UniqueItemID, Tuple[ItemID, FrozenSet[TagID]]] = dict()
        self.uiid_by_item: Dict[Tuple[ItemID, FrozenSet[TagID]], UniqueItemID] = dict()
        self.item_tags_by_tag: Dict[TagID, Set[FrozenSet[TagID]]] = dict()
        self.neighbor_tags_of_a_tag: Dict[TagID, Set[TagID]] = dict()
        self.set_of_tags: Set[TagID] = set()
        self.direct_items_by_tags: Dict[FrozenSet[TagID], Set[ItemID]] = dict()
        self.tags_by_number_of_tags: Dict[int, Set[FrozenSet[TagID]]] = dict()
        
        # self.tags_by_number_of_tags_by_tag: Dict[TagID, Dict[int, Set[FrozenSet[TagID]]]] = dict()
        self.tags_len_by_tag: Dict[TagID, int] = dict()  # key: tag; value: keyb  from self.tags_by_number_of_tags
        
        self.tags_by_number_of_tags_by_tag_by_near_tag: Dict[TagID, Dict[TagID, Dict[int, Set[FrozenSet[TagID]]]]] = \
            dict()  # Dict[interested tag, Dict[near tag, Dict[len of tags set, Set[tag sets]]]]
        self.tags_len_by_tag_by_near_tag: Dict[TagID, Dict[TagID, int]] = \
            dict()  # Dict[interested tag, Dict[near tag, len of tags set]]
        
        self.tags_inclusion_by_tags: Dict[FrozenSet[TagID], Set[FrozenSet[TagID]]] = dict()  # key: interested tags set;
        # value: bigger tags sets which includes interested tags set
        self.tag_weight_by_tag_by_tags: Dict[FrozenSet[TagID], Dict[TagID, int]] = dict()
        self.tags_by_tag_weight_by_tags: Dict[FrozenSet[TagID], Dict[int, Set[TagID]]] = dict()
        self.nearest_non_empty_tags_by_some_tags: Dict[FrozenSet[TagID], Set[FrozenSet[TagID]]] = dict()  # key: any
        # interested tags set (input from user); value: set of known tags sets which has an item(-s) and includes given
        # interested tags set
        self.tags_by_itself: Dict[FrozenSet[TagID], FrozenSet[TagID]] = dict()  # key: tags; value: same tags object

    def set_scheduler(self, scheduler: Optional[Callable]):
        self.scheduler = scheduler or default_scheduler

    def get_scheduler(self):
        return self.scheduler

    def put(self, item_id: ItemID, tags: FrozenSet[TagID]) -> UniqueItemID:
        unique_item_id: UniqueItemID = self.uiidg()
        item = (item_id, tags)
        tags_len = len(tags)
        
        self.item_by_uiid[unique_item_id] = item
        
        self.uiid_by_item[item] = unique_item_id
        
        for tag in tags:
            # self.item_tags_by_tag
            if tag not in self.item_tags_by_tag:
                self.item_tags_by_tag[tag] = set()
            
            self.item_tags_by_tag[tag].add(tags)
        
            # self.neighbor_tags_of_a_tag
            if tag not in self.neighbor_tags_of_a_tag:
                self.neighbor_tags_of_a_tag[tag] = set()
            
            self.neighbor_tags_of_a_tag[tag].update(tags - tag)
            
            # self.set_of_tags
            self.set_of_tags.add(tag)
        
        # self.direct_items_by_tags
        if tags not in self.direct_items_by_tags:
            self.direct_items_by_tags[tags] = set()
        
        self.direct_items_by_tags[tags].add(item_id)
        
        # self.tags_by_number_of_tags
        if tags_len not in self.tags_by_number_of_tags:
            self.tags_by_number_of_tags[tags_len] = set()
        
        self.tags_by_number_of_tags[tags_len].add(tags)
        
        # self.tags_len_by_tag
        for tag in tags:
            self.tags_len_by_tag[tag] = tags_len

    def get(self, item_id: ItemID, tags: FrozenSet[TagID]) -> Optional[UniqueItemID]:
        return self.uiid_by_item.get((item_id, tags), None)

    def item(self, unique_item_id: UniqueItemID) -> Optional[Tuple[ItemID, FrozenSet[TagID]]]:
        return self.item_by_uiid.get(unique_item_id, None)

    def delete(self, unique_item_id: UniqueItemID):
        pass

    def list_direct_items(self, tags: FrozenSet[TagID]):
        pass

    def list_all_items(self, tags: FrozenSet[TagID]):
        pass

    def list_direct_tags(self, tags: FrozenSet[TagID]):
        pass

    def list_all_tags(self, tags: FrozenSet[TagID]):
        pass

    def get_nearest_non_empty_tags_set(self, tags: FrozenSet[TagID]) -> Optional[FrozenSet[TagID]]:
        return self._gen_nearest_non_empty_tags_by_some_tags(True, self._normalise_tags(self._check_tags(tags)))

    def _gen_nearest_non_empty_tags_by_some_tags(
            self, is_search_request: bool, tags: FrozenSet[TagID]) -> Optional[Set[FrozenSet[TagID]]]:
        if tags in self.tags_by_itself:
            return set(self.tags_by_itself[tags])

        if tags in self.nearest_non_empty_tags_by_some_tags:
            return self.nearest_non_empty_tags_by_some_tags[tags]

        nearest_non_empty_tags_set = set()

        if tags in self.tags_inclusion_by_tags:
            inclusion_tags = self.tags_inclusion_by_tags[tags]
            min_tags_len = len(min(inclusion_tags))
            for tags2 in inclusion_tags:
                if len(tags2) == min_tags_len:
                    nearest_non_empty_tags_set.add(tags2)
        else:
            minimal_nearest_non_empty_tags_size = len(tags) + 1
            tags_by_number_of_tags: Dict[int, Set[FrozenSet[TagID]]] = dict()
            for tag in tags:
                lookup1: Dict[TagID, Dict[int, Set[FrozenSet[TagID]]]] = \
                    self.near_tags_by_number_of_tags_by_tag_by_tag[tag]
                for tag2 in tags:
                    if tag == tag2:
                        continue
                    if tag2 not in lookup1:
                        return None
                    lookup2: Dict[int, Set[FrozenSet[TagID]]] = lookup1[tag2]
                    sorted_tags_sizes = sorted(lookup2.keys())
                    for tags_size in sorted_tags_sizes:
                        if tags_size < minimal_nearest_non_empty_tags_size:
                            continue
                        lookup3 = lookup2[tags_size]
                        for another_tags in lookup3:
                            if tags != (tags & another_tags):
                                continue
                            if tags_size not in tags_by_number_of_tags:
                                tags_by_number_of_tags[tags_size] = set()
                            tags_by_number_of_tags[tags_size].add(another_tags)
            min_tags_len = min(tags_by_number_of_tags.keys())
            nearest_non_empty_tags_set = tags_by_number_of_tags[min_tags_len]

            if self._is_growth_allowed(is_search_request):
                self.tags_inclusion_by_tags[tags] = set()
                for inclusion_tags_set in tags_by_number_of_tags.values():
                    self.tags_inclusion_by_tags[tags].update(inclusion_tags_set)

        if self._is_growth_allowed(is_search_request):
            self.nearest_non_empty_tags_by_some_tags[tags] = nearest_non_empty_tags_set

        return nearest_non_empty_tags_set

    def _normalise_tags(self, tags: FrozenSet[TagID]) -> FrozenSet[TagID]:
        if tags in self.tags_by_itself:
            return self.tags_by_itself[tags]
        else:
            return tags

    def _check_tags(self, tags: FrozenSet[TagID]) -> FrozenSet[TagID]:
        if not (tags == (tags & self.set_of_tags)):
            raise TagNotFound(tags - (tags & self.set_of_tags))
        return tags

    def _is_growth_allowed(self, is_search_request: bool):
        return (not is_search_request) or (is_search_request and self.growth_during_the_search_is_allowed())



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# import pip
# from pip._internal import main as pip_main

from cengal.os.execute import prepare_py_params
from cengal.cross_version.console_print.universal import cross_print

import platform
import sys
import os.path
from contextlib import contextmanager

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


PLATFORM_NAME = platform.python_implementation()  # can be 'PyPy', 'CPython', etc.
PYTHON_VERSION = platform.python_version_tuple()  # tuple() of str(); for example: ('3', '5', '1')
PYTHON_VERSION_INT = sys.version_info  # sys.version_info == (major=3, minor=2, micro=5, releaselevel='final', serial=0)
#   Usage: sys.version_info >= (3, 7)


# On Python < 3.4 you should install enum backport module before you will be able to use it. So:
FROM_PYPI = 0
FROM_ZIP = 1
FROM_FOLDER_WITH_ZIP_FILES = 2
FROM_EXTERNAL_GIT = 3


_FILTERING_TYPE__OFF = 0
_FILTERING_TYPE__INCLUDING = 1
_FILTERING_TYPE__EXCLUDING = 2


# Do not import this function from THIS module!
@contextmanager
def _change_current_dir(new_current_dir):
    cur_dir = os.getcwd()
    os.chdir(new_current_dir)
    try:
        yield
    except:
        raise
    finally:
        os.chdir(cur_dir)


# Do not import this function from THIS module!
def _filtered_file_list_traversal(root_dir, filtering_type, extentions_set=None, remove_empty_items=False):
    """
    :param root_dir: str(); r'C:\dir\path'
    :param filtering_type: FilteringType()
    :param extentions_set: set(); {'.upk', '.txt'}
    :return: list() of tuple(); list of (dirpath, dirnames, new_filenames)
    """

    if (_FILTERING_TYPE__OFF != filtering_type) and (extentions_set is None):
        raise Exception('extentions_set can\'t be None with this filtering type')

    result_list = list()
    for raw_result in os.walk(root_dir):
        dirpath = raw_result[0]
        dirnames = raw_result[1]
        filenames = raw_result[2]

        new_filenames = list()
        for filename in filenames:
            if _FILTERING_TYPE__OFF == filtering_type:
                new_filenames.append(filename)
            elif _FILTERING_TYPE__INCLUDING == filtering_type:
                extention = os.path.splitext(filename)[1]
                if extention in extentions_set:
                    new_filenames.append(filename)
            elif _FILTERING_TYPE__EXCLUDING == filtering_type:
                extention = os.path.splitext(filename)[1]
                if extention not in extentions_set:
                    new_filenames.append(filename)
        if remove_empty_items:
            if len(new_filenames) > 0:
                result = (dirpath, dirnames, new_filenames)
                result_list.append(result)
        else:
            result = (dirpath, dirnames, new_filenames)
            result_list.append(result)

    return result_list


# def pip_install(name_or_path):
#     command_line = None
#     pip_name = None
#     if 'nt' == os.name:
#         path_to_interpreter = sys.executable
#         interpreter_root_dir_name = os.path.dirname(path_to_interpreter)
#         path_to_pip = os.path.join(interpreter_root_dir_name, 'Scripts', 'pip.exe')
#         pip_name_template = '"{}"'
#         pip_name = pip_name_template.format(path_to_pip)
#     else:
#         if '2' == PYTHON_VERSION[0]:
#             pip_name = 'pip'
#         elif '3' == PYTHON_VERSION[0]:
#             pip_name = 'pip3'
#     command_line_template = '{pip_name} install --upgrade {package_name_or_path}'
#     command_line = command_line_template.format(pip_name=pip_name, package_name_or_path=name_or_path)
#     os.system(command_line)


def pip_install(name_or_path):
    os.system(prepare_py_params(['-m', 'pip', 'install', '--upgrade', name_or_path]))


def install(name_or_path, type_of_installation):
    """
    Will exit(1) on error (done by pip_main() itself)!
    :param name_or_path: 'tornado', './python-lzf-master-python3.zip', './InstallSourcess/Python3/', etc.
    :param type_of_installation: {FROM_PYPI, FROM_ZIP, FROM_FOLDER_WITH_ZIP_FILES}
    :return:
    """
    if FROM_PYPI == type_of_installation:
        # pip_main(['install', name_or_path])
        pip_install(name_or_path)
        cross_print('')
    elif FROM_ZIP == type_of_installation:
        # pip_main(['install', name_or_path])
        pip_install(name_or_path)
        cross_print('')
    elif FROM_FOLDER_WITH_ZIP_FILES == type_of_installation:
        interested_extensions = {'.zip'}
        result_list = _filtered_file_list_traversal(name_or_path, _FILTERING_TYPE__INCLUDING, interested_extensions)
        for result in result_list:
            for filename in result[2]:
                # I'm not sure now, but I think I remember that there was some troubles when you trying to install
                # some zipped modules with full path under PyPy. So:
                relative_file_name = os.path.join('.', filename)
                with _change_current_dir(result[0]):
                    # pip_main(['install', relative_file_name])
                    pip_install(relative_file_name)
                    cross_print('')



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['Lmdb', 'LmdbRequest', 'KeyType', 'RawKeyType', 'ValueType', 'RawValueType', 'DbId', 'DbName', 'DbKeyError']

from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_tools.await_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.asyncio_loop import *
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import CoroPriority
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.timer_func_runner import *
from cengal.file_system.file_manager import path_relative_to_current_dir
from cengal.time_management.cpu_clock_cycles import perf_counter
from cengal.data_manipulation.serialization import *
from cengal.introspection.inspect import get_exception
from typing import Hashable, Tuple, List, Any, Dict, Callable
import sys
import os
import asyncio
try:
    import lmdb as lmdb_lib
except ImportError:
    from warnings import warn
    warn('''WARNING: `lmdb` library is not installed. Lmdb service will not work.
         To install `lmdb` use: `pip install lmdb`''')
    raise


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


KeyType = Union[bytes, str, Any]
RawKeyType = bytes
ValueType = Any
RawValueType = bytes
DbId = Hashable
DbName = bytes


class LmdbRequest(ServiceRequest):
    def set_db_environment_path(self, path_to_db_environment: str) -> ServiceRequest:
        return self._save(0, path_to_db_environment)
    
    def open_databases(self, db_names: Dict[DbId, DbName]) -> ServiceRequest:
        return self._save(1, db_names)
    
    def drop_db(self, db_id: DbId, delete: bool = False) -> ServiceRequest:
        return self._save(2, db_id, delete)
    
    def sync(self) -> ServiceRequest:
        return self._save(3)
    
    def get(self, key: KeyType, db_id: DbId = None) -> ServiceRequest:
        return self._save(4, key, db_id)
    
    def get_items(self, keys: Set[Tuple[KeyType, DbId]]) -> ServiceRequest:
        return self._save(5, keys)
    
    def get_all_items(self, db_id: DbId = None) -> ServiceRequest:
        return self._save(6, db_id)
    
    def put(self, key: KeyType, value: Optional[ValueType] = None, db_id: DbId = None) -> ServiceRequest:
        return self._save(7, key, value, db_id)
    
    def put_items(self, items: Dict[Tuple[KeyType, DbId], Optional[ValueType]]) -> ServiceRequest:
        return self._save(8, items)
    
    def delete(self, key: KeyType, value: Optional[ValueType] = None, db_id: DbId = None) -> ServiceRequest:
        return self._save(9, key, value, db_id)
    
    def delete_items(self, items: Dict[Tuple[KeyType, DbId], Optional[ValueType]]) -> ServiceRequest:
        return self._save(10, items)

    def open_db_environment(self, path_to_db_environment: str) -> ServiceRequest:
        return self._save(11, path_to_db_environment)
    

def make_key_frozen(key):
    if isinstance(key, list) or isinstance(key, set):
        new_key = list()
        for item in key:
            new_key.append(make_key_frozen(item))
        
        key = tuple(new_key)

    return key
    

class Lmdb(Service, EntityStatsMixin):
    def __init__(self, loop: CoroSchedulerType):
        super(Lmdb, self).__init__(loop)
        self.default_db_name: DbName = b'__default__'
        self.drop_db_requests: Dict[CoroID, Tuple[Hashable, bool]] = dict()
        self.read_queue: List[Tuple[CoroID, Tuple[RawKeyType, DbId]]] = list()
        self.massive_read_queue: List[Tuple[CoroID, Set[Tuple[KeyType, DbId]]]] = list()
        self.data_cache: Dict[Tuple[Hashable, Hashable], Any] = dict()
        self.deletion_cache: Dict[Tuple[Hashable, Hashable], Any] = dict()
        self.get_all_items_queue: List[Tuple[CoroID, DbId]] = list()
        self.path_to_db_environment = path_relative_to_current_dir('lmdb.db')
        self.db_environment = None
        self.databases: Dict[Hashable, Any] = dict()
        self.db_names: Dict[DbId, DbName] = dict()
        self.async_loop = None
        self.sync_time_interval = 1.0
        self.last_sync_time = perf_counter()
        self.force_sync = False
        self.write_locked = False
        self.writes_num: int = 0
        self.reads_num: int = 0
        self.deletes_num: int = 0
        self.db_drops_num: int = 0
        self.write_locked_coro_id: Optional[CoroID] = None
        # self.serializer = best_serializer_for_standard_data((DataFormats.binary,
        #                                    Tags.can_use_bytes,
        #                                    Tags.decode_str_as_str,
        #                                    Tags.decode_list_as_list,
        #                                    Tags.decode_bytes_as_bytes,
        #                                    Tags.superficial,
        #                                    Tags.current_platform,
        #                                    Tags.multi_platform),
        #                                   TestDataType.small,
        #                                   0.1)
        self.serializer = Serializer(Serializers.msgspec_messagepack)

        self._request_workers = {
            0: self._on_set_db_environment_path,
            1: self._open_databases,
            2: self._drop_db,
            3: self._on_sync,
            4: self._on_get,
            5: self._on_get_items,
            6: self._on_get_all_items,
            7: self._on_put,
            8: self._on_put_items,
            9: self._on_delete,
            10: self._on_delete_items,
            11: self._open_db_environment,
        }

    def get_entity_stats(self, stats_level: 'EntityStatsMixin.StatsLevel' = EntityStatsMixin.StatsLevel.debug) -> Tuple[str, Dict[str, Any]]:
        return type(self).__name__, {
            'db_names': list(self.db_names.keys()),
            'writes num': self.writes_num,
            'reads num': self.reads_num,
            'deletes num': self.deletes_num,
            'db drops num': self.db_drops_num,
        }

    def single_task_registration_or_immediate_processing(
            self, *args, **kwargs) -> ServiceProcessingResponse:
        result = self.try_resolve_request(*args, **kwargs)
        if result is None:
            return True, None, None
        else:
            return result

    def full_processing_iteration(self):
        self.force_sync = False
        
        if self.db_environment is None:
            self._init_db()

        data_cache_buff = self.data_cache
        self.data_cache = type(data_cache_buff)()
        
        read_queue_buff = self.read_queue
        self.read_queue = type(read_queue_buff)()
        
        massive_read_queue_buff = self.massive_read_queue
        self.massive_read_queue = type(massive_read_queue_buff)()
        
        deletion_cache_buff = self.deletion_cache
        self.deletion_cache = type(deletion_cache_buff)()
        
        get_all_items_queue_buff = self.get_all_items_queue
        self.get_all_items_queue = type(get_all_items_queue_buff)()

        # put
        def put_handler(db_environment, databases):
            db_id = None
            try:
                with db_environment.begin(write=True) as txn:
                    for key_info, value in data_cache_buff.items():
                        key, db_id = key_info
                        txn.put(key, value, db=databases[db_id], dupdata=False, append=False)

                self.writes_num += len(data_cache_buff)
            except lmdb_lib.MapFullError:
                raise DBError.from_exception(db_id)
        
        lmdb_reapplier(self.db_environment, self.databases, put_handler)
        
        # delete
        for key_info, value in deletion_cache_buff.items():
            with self.db_environment.begin(write=True) as txn:
                key, db_id = key_info
                txn.delete(key, value, db=self.databases[db_id])
                self.deletes_num += 1

        # drop
        drop_db_requests_buff = self.drop_db_requests
        self.drop_db_requests = type(drop_db_requests_buff)()
        dropped_databases: Set[Hashable] = set()
        processed_coroutines: Set[CoroID] = set()
        
        def drop_handler(db_environment, databases):
            for coro_id, request in drop_db_requests_buff.items():
                if coro_id in processed_coroutines:
                    continue
                
                db_id, delete_db = request
                if db_id not in dropped_databases:
                    try:
                        with db_environment.begin(write=True) as txn:
                            txn.drop(db=databases[db_id], delete=delete_db)
                        
                        self.db_drops_num += 1
                    except lmdb_lib.MapFullError:
                        raise DBError.from_exception(db_id)
                    
                    dropped_databases.add(db_id)
                
                self.register_response(coro_id, None, None)
                processed_coroutines.add(coro_id)
                    
        lmdb_reapplier(self.db_environment, self.databases, drop_handler)

        # get
        def get_item(txn, key_info, data_cache_buff) -> Tuple[ValueType, Optional[Exception]]:
            key, db_id = key_info
            if key_info in data_cache_buff:
                value = data_cache_buff[key_info]
            else:
                value = txn.get(key, db=self.databases[db_id])
                self.reads_num += 1
            
            exception = None
            try:
                if value is None:
                    exception = DbKeyError(key_info)
                else:
                    value = self.serializer.loads(value)
            except:
                exception = get_exception()
            
            return value, exception
            
        with self.db_environment.begin() as txn:
            for coro_id, key_info in read_queue_buff:
                value, exception = get_item(txn, key_info, data_cache_buff)
                self.register_response(coro_id, value, exception)
            
            for coro_id, set_of_key_info in massive_read_queue_buff:
                items: Dict[Tuple[KeyType, DbId], Tuple[ValueType, Optional[Exception]]] = dict()
                for key_info in set_of_key_info:
                    items[key_info] = get_item(txn, key_info, data_cache_buff)
                    self.register_response(coro_id, items, None)

        # get all items
        for coro_id, db_id in get_all_items_queue_buff:
            with self.db_environment.begin(db=self.databases[db_id]) as txn:
                result = dict()
                exception = None
                try:
                    # for k, v in txn.cursor(db=self.databases[db_id]):
                    #     key = make_key_frozen(self.serializer.loads(k))
                    #     value = self.serializer.loads(v)
                    #     result[key] = value
                    result = {make_key_frozen(self.serializer.loads(k)): self.serializer.loads(v) for k, v in txn.cursor(db=self.databases[db_id])}
                    self.reads_num += len(result)
                except:
                    exception = get_exception()
                
                self.register_response(coro_id, result, exception)
        
        # sync
        self.sync_in_thread_pool()
        
        self.last_sync_time = perf_counter()

        self.make_dead()

    def in_work(self) -> bool:
        result: bool = (bool(self.read_queue) or bool(self.massive_read_queue) or bool(self.get_all_items_queue)) or (self.force_sync or ((not bool(self.write_locked)) and (bool(self.data_cache) or bool(self.deletion_cache) or bool(self.drop_db_requests)) and ((perf_counter() - self.last_sync_time) >= self.sync_time_interval)))
        return self.thrifty_in_work(result)
    
    def time_left_before_next_event(self) -> Tuple[bool, Optional[Union[int, float]]]:
        time_since_last_sync_time: float = perf_counter() - self.last_sync_time
        if self.sync_time_interval > time_since_last_sync_time:
            return True, self.sync_time_interval - time_since_last_sync_time
        else:
            return True, 0

    def _init_db(self):
        print(f'self.path_to_db_environment: {self.path_to_db_environment}')
        self.db_environment = lmdb_lib.open(self.path_to_db_environment, map_size=20 * 1024**2, writemap=True, max_dbs=10,
                                        map_async=True, lock=False, metasync=False, sync=False, meminit=False)
        self.databases[None] = self.db_environment.open_db(self.default_db_name)
        self.db_environment.sync(True)
    
    def _open_db_environment(self, path_to_db_environment: str) -> ServiceProcessingResponse:
        raise NotImplementedError

    def _on_set_db_environment_path(self, path_to_db_environment: str) -> ServiceProcessingResponse:
        if self.write_locked:
            return True, False, None
        
        if self.db_environment is None:
            self.path_to_db_environment = path_to_db_environment
            try:
                self._init_db()
            except:
                exception = get_exception()
                return True, False, exception
            return True, True, None
        else:
            return True, False, None
    
    def _on_sync(self) -> ServiceProcessingResponse:
        if self.data_cache:
            self.force_sync = True
            self.make_live()
        else:
            # self.db_environment.sync(True)
            self.sync_in_thread_pool()
        
        return True, None, None
    
    def _on_get(self, key: KeyType, db_id: DbId = None) -> ServiceProcessingResponse:
        key = self.serializer.dumps(key)
        
        key_info = (key, db_id)
        if key_info in self.data_cache:
            return True, self.data_cache[key_info], None
        else:
            self.read_queue.append((self.current_caller_coro_info.coro_id, key_info))
            self.make_live()
            return False, None, None
    
    def _on_get_items(self, keys: Set[Tuple[KeyType, DbId]]) -> ServiceProcessingResponse:
        coro_id = self.current_caller_coro_info.coro_id
        
        raw_keys: Set[Tuple[KeyType, DbId]] = set()
        for key, db_id in keys:
            key = self.serializer.dumps(key)
            
            raw_keys.add((key, db_id))
        
        self.massive_read_queue.append((coro_id, raw_keys))
        self.make_live()
        return False, None, None
    
    def _on_get_all_items(self, db_id: DbId) -> ServiceProcessingResponse:
        coro_id = self.current_caller_coro_info.coro_id
        self.get_all_items_queue.append((coro_id, db_id))
        self.make_live()
        return False, None, None
    
    def _on_put(self, key: KeyType, value: Any, db_id: DbId = None) -> ServiceProcessingResponse:
        key = self.serializer.dumps(key)
        
        key_info = (key, db_id)
        exception = None
        result = None
        try:
            result = self.data_cache[key_info] = self.serializer.dumps(value)
        except:
            exception = get_exception()
        
        self.make_live()
        return True, result, exception
    
    def _on_put_items(self, items: Dict[Tuple[KeyType, DbId], ValueType]) -> ServiceProcessingResponse:
        result_items: Dict[Tuple[KeyType, DbId], Tuple[ValueType, Optional[Exception]]] = dict()
        for key_info, value in items.items():
            key, db_id = key_info
            key = self.serializer.dumps(key)
            
            key_info = (key, db_id)
            exception = None
            result = None
            try:
                result = self.data_cache[key_info] = self.serializer.dumps(value)
            except:
                exception = get_exception()
            
            result_items[key_info] = (result, exception)
        
        self.make_live()
        return True, result_items, None
    
    def _on_delete(self, key: KeyType, value: Any, db_id: DbId = None) -> ServiceProcessingResponse:
        key = self.serializer.dumps(key)
        
        key_info = (key, db_id)
        exception = None
        result = None
        try:
            result = self.deletion_cache[key_info] = self.serializer.dumps(value)
        except:
            exception = get_exception()
        
        self.make_live()
        return True, result, exception
    
    def _on_delete_items(self, items: Dict[Tuple[KeyType, DbId], ValueType]) -> ServiceProcessingResponse:
        result_items: Dict[Tuple[KeyType, DbId], Tuple[ValueType, Optional[Exception]]] = dict()
        for key_info, value in items.items():
            key, db_id = key_info
            key = self.serializer.dumps(key)
            
            key_info = (key, db_id)
            exception = None
            result = None
            try:
                result = self.deletion_cache[key_info] = self.serializer.dumps(value)
            except:
                exception = get_exception()
            
            result_items[key_info] = (result, exception)
        
        self.make_live()
        return True, result_items, None
    
    def _open_databases(self, db_names: Dict[DbId, DbName]) -> ServiceProcessingResponse:
        for db_id, db_name in db_names.items():
            self.databases[db_id] = self.db_environment.open_db(db_name)
            self.db_names[db_id] = db_name
        
        self.db_environment.sync(True)
        return True, None, None
    
    def _drop_db(self, db_id: DbId, delete: bool = False) -> ServiceProcessingResponse:
        coro_id = self.current_caller_coro_info.coro_id
        self.drop_db_requests[coro_id] = (db_id, delete)
        self.make_live()
        return False, None, None
    
    def sync_in_thread_pool(self):
        async def sync_db_coro(i: Interface, self, asyncio_loop, need_to_ensure_asyncio_loop: bool):
            if need_to_ensure_asyncio_loop:
                asyncio_loop = await i(AsyncioLoop, AsyncioLoopRequest().ensure_loop(None,CoroPriority.low, True))
            else:
                if asyncio_loop is None:
                    asyncio_loop = await i(AsyncioLoop, AsyncioLoopRequest().get())
            
            async def sync_db(self, asyncio_loop):
                def sync_worker():
                    self.db_environment.sync(True)
                
                await task_in_thread_pool(asyncio_loop, sync_worker)

            await i(AsyncioLoop, AsyncioLoopRequest().wait(sync_db(self, asyncio_loop)))
            self.write_locked_coro_id = None
            self.write_locked = False
            def make_service_live_for_a_next_sync(self):
                self.make_live()
            
            await i(TimerFuncRunner, self.sync_time_interval, make_service_live_for_a_next_sync, self)

        asyncio_loop = None
        need_to_ensure_asyncio_loop = False
        try:
            asyncio_loop = self._loop.get_service_instance(AsyncioLoop).inline_get()
        except AsyncioLoopWasNotSetError:
            need_to_ensure_asyncio_loop = True

        coro: CoroWrapperBase = self._loop.put_coro(sync_db_coro, self, asyncio_loop, need_to_ensure_asyncio_loop)
        self.write_locked = True
        self.write_locked_coro_id = coro.coro_id


LmdbRequest.default_service_type = Lmdb


class DbKeyError(KeyError):
    def __init__(self, key_info: Tuple[KeyType, DbId], *args: object) -> None:
        super().__init__(*args)
        self.key_info: Tuple[KeyType, DbId] = key_info


class DBError(Exception):
    def __init__(self, db_id: DbId, original_exception: Exception, *args):
        super().__init__(*args)
        self.db_id: DbId = db_id
        self.original_exception = original_exception
    
    @staticmethod
    def from_exception(db_id: DbId) -> 'DBError':
        return DBError(db_id, get_exception())


def lmdb_reapplier(environment: lmdb_lib.Environment, databases: Dict[Hashable, Any], handler: Callable):
    failed = True
    while failed:
        need_to_resize: bool = False
        try:
            handler(environment, databases)
            failed = False
        except DBError as err:
            if isinstance(err.original_exception, lmdb_lib.MapFullError):
                need_to_resize = True
        
        if need_to_resize:
            environment.set_mapsize(environment.info()['map_size'] + 2 * 1024**2)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
from contextlib import contextmanager, redirect_stdout, redirect_stderr
from cengal.system import IS_RUNNING_IN_PYCHARM
if 'nt' == os.name:
    import win32console
    import codecs


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


# if 'nt' == os.name:
#     @contextmanager
#     def change_console_output_cp(new_output_cp=None, new_input_cp=None, encoding_error=None):
#         """
#         :param new_output_cp: ResultExistence(True, (65001, 'utf-8))
#         :param new_input_cp: ResultExistence(True, (65001, 'utf-8))
#         :return:
#         """
#         encoding_error = encoding_error or 'replace'
#         cur_output_cp = None
#         cur_input_cp = None
#         cur_output_cp_name = str()
#         cur_input_cp_name = str()
#         original_stdout = sys.stdout
#         original_stderr = sys.stderr
#         original_stdin = sys.stdin
#         original_stdout_fileno = sys.stdout.fileno()
#         original_stderr_fileno = sys.stderr.fileno()
#         original_stdin_fileno = sys.stdin.fileno()
#         if 'nt' == os.name:
#             if not IS_RUNNING_IN_PYCHARM:
#                 cur_output_cp = win32console.GetConsoleOutputCP()
#                 cur_output_cp_name = sys.stdout.encoding
#                 if (new_output_cp is not None) and new_output_cp.existence:
#                     # sys.stdout = codecs.getwriter(new_output_cp.result[1])(sys.stdout.buffer, encoding_error)
#                     # sys.stderr = codecs.getwriter(new_output_cp.result[1])(sys.stderr.buffer, encoding_error)
#                     # sys.stdout = codecs.getwriter(new_output_cp.result[1])(sys.stdout.detach(), encoding_error)
#                     # sys.stderr = codecs.getwriter(new_output_cp.result[1])(sys.stderr.detach(), encoding_error)
#                     sys.stdout = open(original_stdout_fileno, mode='w', encoding=new_output_cp.result[1], buffering=1)
#                     print('stdout:', sys.stdout.encoding)
#                     sys.stderr = open(original_stderr_fileno, mode='w', encoding=new_output_cp.result[1], buffering=1)
#                     print('stderr:', sys.stderr.encoding)
#                     win32console.SetConsoleOutputCP(new_output_cp.result[0])
#                     print('conout:', win32console.GetConsoleOutputCP())
#                 cur_input_cp = win32console.GetConsoleCP()
#                 cur_input_cp_name = sys.stdin.encoding
#                 if (new_input_cp is not None) and new_input_cp.existence:
#                     # sys.stdin = codecs.getwriter(new_input_cp.result[1])(sys.stdin.buffer, encoding_error)
#                     sys.stdin = open(original_stdin_fileno, mode='r', encoding=new_input_cp.result[1], buffering=1)
#                     print('stdin:', sys.stdin.encoding)
#                     win32console.SetConsoleCP(new_input_cp.result[0])
#                     print('conin:', win32console.GetConsoleOutputCP())
#         try:
#             yield ((cur_input_cp, cur_output_cp), (cur_input_cp_name, cur_output_cp_name))
#         except:
#             raise
#         finally:
#             if 'nt' == os.name:
#                 if not IS_RUNNING_IN_PYCHARM:
#                     if (new_output_cp is not None) and new_output_cp.existence:
#                         win32console.SetConsoleOutputCP(cur_output_cp)
#                         print('conout:', win32console.GetConsoleOutputCP())
#                         # sys.stdout.close()
#                         # sys.stdout = original_stdout
#                         # sys.stderr = original_stderr
#                         # sys.stdout = open(original_stdout_fileno, mode='w', encoding=cur_output_cp_name, buffering=1)
#                         print('stdout:', sys.stdout.encoding)
#                         # sys.stderr.close()
#                         # sys.stderr = open(original_stderr_fileno, mode='w', encoding=cur_output_cp_name, buffering=1)
#                         print('stderr:', sys.stderr.encoding)
#                     if (new_input_cp is not None) and new_input_cp.existence:
#                         win32console.SetConsoleCP(cur_input_cp)
#                         print('conin:', win32console.GetConsoleOutputCP())
#                         # sys.stdin.close()
#                         # sys.stdin = original_stdin
#                         # sys.stdin = open(original_stdin_fileno, mode='r', encoding=cur_input_cp_name, buffering=1)
#                         print('stdin:', sys.stdin.encoding)
# else:
#     @contextmanager
#     def change_console_output_cp(new_output_cp=None, new_input_cp=None, encoding_error=None):
#         yield None


def fileno(file_or_fd):
    fd = getattr(file_or_fd, 'fileno', lambda: file_or_fd)()
    if not isinstance(fd, int):
        raise ValueError("Expected a file (`.fileno()`) or a file descriptor")
    return fd


if 'nt' == os.name:
    @contextmanager
    def change_console_output_cp(new_output_cp=None, new_input_cp=None, encoding_error=None):
        """
        :param new_output_cp: ResultExistence(True, (65001, 'utf-8))
        :param new_input_cp: ResultExistence(True, (65001, 'utf-8))
        :return:
        """
        encoding_error = encoding_error or 'replace'
        cur_output_cp = None
        cur_input_cp = None
        cur_output_cp_name = str()
        cur_input_cp_name = str()
        original_stdout = sys.stdout
        original_stderr = sys.stderr
        original_stdin = sys.stdin
        original_stdout_fileno = sys.stdout.fileno()
        original_stderr_fileno = sys.stderr.fileno()
        original_stdin_fileno = sys.stdin.fileno()
        new_stdout = None
        new_stderr = None
        new_stdin = None
        if 'nt' == os.name:
            if IS_RUNNING_IN_PYCHARM:
                yield None
            else:
                with os.fdopen(os.dup(original_stdout_fileno), 'wb') as copied_stdout, \
                        os.fdopen(os.dup(original_stderr_fileno), 'wb') as copied_stderr, \
                        os.fdopen(os.dup(original_stdin_fileno), 'rb') as copied_stdin, \
                        os.fdopen(os.dup(original_stdout_fileno), 'wb') as copied_stdout_buff, \
                        os.fdopen(os.dup(original_stderr_fileno), 'wb') as copied_stderr_buff, \
                        os.fdopen(os.dup(original_stdin_fileno), 'rb') as copied_stdin_buff:
                    # copied_stdout = os.fdopen(os.dup(original_stdout_fileno), 'wb')
                    # copied_stderr = os.fdopen(os.dup(original_stderr_fileno), 'wb')
                    # copied_stdin = os.fdopen(os.dup(original_stdin_fileno), 'wb')
                    # copied_stdout_buff = os.fdopen(os.dup(original_stdout_fileno), 'wb')
                    # copied_stderr_buff = os.fdopen(os.dup(original_stderr_fileno), 'wb')
                    # copied_stdin_buff = os.fdopen(os.dup(original_stdin_fileno), 'wb')
                    original_stdout.flush()
                    original_stderr.flush()
                    original_stdin.flush()

                    cur_output_cp = win32console.GetConsoleOutputCP()
                    cur_output_cp_name = sys.stdout.encoding
                    if (new_output_cp is not None) and new_output_cp.existence:
                        new_stdout = open(copied_stdout.fileno(), mode='w', encoding=new_output_cp.result[1],
                                          buffering=1)
                        sys.stdout = new_stdout
                        # print('stdout:', sys.stdout.encoding)
                        new_stderr = open(copied_stderr.fileno(), mode='w', encoding=new_output_cp.result[1],
                                          buffering=1)
                        sys.stderr = new_stderr
                        # print('stderr:', sys.stderr.encoding)
                        win32console.SetConsoleOutputCP(new_output_cp.result[0])
                        # print('conout:', win32console.GetConsoleOutputCP())
                    cur_input_cp = win32console.GetConsoleCP()
                    cur_input_cp_name = sys.stdin.encoding
                    if (new_input_cp is not None) and new_input_cp.existence:
                        new_stdin = open(copied_stdin.fileno(), mode='r', encoding=new_input_cp.result[1],
                                         buffering=1)
                        sys.stdin = new_stdin
                        # print('stdin:', sys.stdin.encoding)
                        win32console.SetConsoleCP(new_input_cp.result[0])
                        # print('conin:', win32console.GetConsoleOutputCP())
                    try:
                        yield ((cur_input_cp, cur_output_cp), (cur_input_cp_name, cur_output_cp_name))
                    except:
                        raise
                    finally:
                        sys.stdout.flush()
                        sys.stderr.flush()
                        sys.stdin.flush()

                        if (new_output_cp is not None) and new_output_cp.existence:
                            win32console.SetConsoleOutputCP(cur_output_cp)
                            # print('conout:', win32console.GetConsoleOutputCP())
                            os.dup2(copied_stdout_buff.fileno(), original_stdout_fileno)
                            sys.stdout = original_stdout
                            # print('stdout:', sys.stdout.encoding)
                            os.dup2(copied_stderr_buff.fileno(), original_stderr_fileno)
                            sys.stderr = original_stderr
                            # print('stderr:', sys.stderr.encoding)
                        if (new_input_cp is not None) and new_input_cp.existence:
                            win32console.SetConsoleCP(cur_input_cp)
                            # print('conin:', win32console.GetConsoleOutputCP())
                            os.dup2(copied_stdin_buff.fileno(), original_stdin_fileno)
                            sys.stdin = original_stdin
                            # print('stdin:', sys.stdin.encoding)
        else:
            yield None
else:
    @contextmanager
    def change_console_output_cp(new_output_cp=None, new_input_cp=None, encoding_error=None):
        yield None



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['TimerCoroRunner', 'add_timer_coro_run_from_other_service', 'discard_timer_coro_run_from_other_service', 'timer_coro_run_on', 'try_timer_coro_run_on', 'atimer_coro_run_on', 'atry_timer_coro_run_on', 'timer_coro_run', 'try_timer_coro_run', 'atimer_coro_run', 'atry_timer_coro_run']

from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services_internal_lib.service_with_a_direct_request import *
from cengal.code_flow_control.smart_values import ValueExistence
from cengal.time_management.timer import Timer, TimerRequest
from functools import partial
from typing import Tuple, Optional, Union
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import put_current_from_other_service, put_root_from_other_service, put_from_other_service


class TimerCoroRunnerRequest(ServiceRequest):
    def add(self, delay: float, coro_worker: AnyWorker, *args, **kwargs) -> TimerRequest:
        return self._save(0, delay, coro_worker, args, kwargs)
    def discard(self, timer_request: TimerRequest) -> bool:
        return self._save(1, timer_request)


class TimerCoroRunner(DualImmediateProcessingServiceMixin, ServiceWithADirectRequestMixin, TypedService[TimerRequest]):
    def __init__(self, loop: CoroSchedulerType):
        super(TimerCoroRunner, self).__init__(loop)
        self.timer = Timer()
        self.pending_tasks_number = 0
        self.direct_requests = list()
        self._request_workers = {
            0: self._on_add,
            1: self._on_discard,
        }

    def single_task_registration_or_immediate_processing_single(
            self, delay: float, coro_worker: AnyWorker, *args, **kwargs) -> Tuple[bool, None]:
        timer_request: TimerRequest = self._add_request_impl(delay, coro_worker, *args, **kwargs)
        self.make_live()
        return True, timer_request, None
    
    def _add_request_impl(self, delay: float, coro_worker: AnyWorker, *args, **kwargs):
        def timer_handler_func(caller_coro_id, coro_worker_: AnyWorker, *args_, **kwargs_):
            put_from_other_service(self, caller_coro_id, coro_worker_, *args_, **kwargs_)
            self.task_triggered()

        timer_handler = partial(timer_handler_func, self.current_caller_coro_info.coro_id, coro_worker, *args, **kwargs)
        self.task_added()
        return self.timer.register(timer_handler, delay)
    
    def _add_request_external_impl(self, delay: float, coro_worker: AnyWorker, *args, **kwargs):
        def timer_handler_func(coro_worker_: AnyWorker, *args_, **kwargs_):
            try:
                put_root_from_other_service(self, coro_worker_, *args_, **kwargs_)
            finally:
                self.task_triggered()

        timer_handler = partial(timer_handler_func, coro_worker, *args, **kwargs)
        self.task_added()
        return self.timer.register(timer_handler, delay)
    
    def _on_add(self, delay: float, coro_worker: AnyWorker, *args, **kwargs) -> Tuple[bool, TimerRequest, None]:
        timer_request: TimerRequest = self._add_request_impl(delay, coro_worker, *args, **kwargs)
        self.make_live()
        return True, timer_request, None
    
    def _on_discard(self, timer_request: TimerRequest) -> Tuple[bool, TimerRequest, None]:
        result: bool = self.timer.discard(timer_request)
        if result:
            self.task_triggered()
        
        return True, result, None
    
    def add_timer_coro_run_from_other_service(self, delay: float, coro_worker: AnyWorker, *args, **kwargs) -> TimerRequest:
        timer_request: TimerRequest = self._add_request_external_impl(delay, coro_worker, *args, **kwargs)
        self.make_live()
        return timer_request
    
    def discard_timer_coro_run_from_other_service(self, timer_request: TimerRequest):
        result: bool = self.timer.discard(timer_request)
        if result:
            self.task_triggered()
        
        return result

    def full_processing_iteration(self):
        if self.direct_requests:
            direct_requests_buff = self.direct_requests
            self.direct_requests = type(direct_requests_buff)()
            for delay, coro_worker, args, kwargs in direct_requests_buff:
                self._add_request_external_impl(delay, coro_worker, *args, **kwargs)

        self.timer()
        if 0 == self.pending_tasks_number:
            self.make_dead()
    
    def _add_direct_request(self, delay: float, coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[None]:
        self.direct_requests.append((delay, coro_worker, args, kwargs))
        self.make_live()
        return (False, None)

    def task_added(self):
        self.pending_tasks_number += 1

    def task_triggered(self):
        self.pending_tasks_number -= 1

    def in_work(self) -> bool:
        result: bool = (self.pending_tasks_number != 0) or bool(self.direct_requests)
        return self.thrifty_in_work(result)
    
    def time_left_before_next_event(self) -> Tuple[bool, Optional[Union[int, float]]]:
        return True, self.timer.nearest_event()


TimerCoroRunnerRequest.default_service_type = TimerCoroRunner


def add_timer_coro_run_from_other_service(current_service: Service, delay: float, coro_worker: AnyWorker, *args, **kwargs) -> TimerRequest:
    timer_coro_runner: TimerCoroRunner = current_service._loop.get_service_instance(TimerCoroRunner)
    return timer_coro_runner.add_timer_coro_run_from_other_service(delay, coro_worker, *args, **kwargs)


def discard_timer_coro_run_from_other_service(current_service: Service, timer_request: TimerRequest) -> bool:
    timer_coro_runner: TimerCoroRunner = current_service._loop.get_service_instance(TimerCoroRunner)
    return timer_coro_runner.discard_timer_coro_run_from_other_service(timer_request)


def timer_coro_run_on(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], delay: float, coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[Optional[CoroID]]:
    """_summary_
        context can be generated by one of the [interface_and_loop_with_backup_loop, get_interface_and_loop_with_backup_loop, interface_and_loop_with_explicit_loop, get_interface_and_loop_with_explicit_loop, interface_for_an_explicit_loop, get_interface_for_an_explicit_loop] functions from the cengal/parallel_execution/coroutines/coro_scheduler module
        
        An example:

        from cengal.parallel_execution.coroutines.coro_scheduler import get_interface_and_loop_with_explicit_loop, CoroSchedulerType, ExplicitWorker, Worker, CoroID
        from cengal.parallel_execution.coroutines.coro_standard_services.timer_coro_runner import timer_coro_run_on
        from typing import Optional, Union

        def my_func(loop: CoroSchedulerType, coro_worker: AnyWorker, a, b) -> Optional[CoroID]:
            try:
                def print_hello_world(i: Interface, name: str):
                    print(f'Hello Wrold from {name}!)
                
                timer_coro_run_on(10, print_hello_world, 'John Doe')
            except CoroSchedulerContextIsNotAvailable:
                print('We are outside of the loop AND no loop was selected as a Primary AND our given `loop` var is None)
        
    Args:
        context (Tuple[Optional[CoroSchedulerType], Optional[Interface], bool]): _description_
        delay (float): delay in seconds
        coro_worker (AnyWorker): coro_worker

    Returns:
        ValueExistence[Optional[CoroID]]: _description_
    """
    return make_request_to_service_with_context(context, TimerCoroRunner, delay, coro_worker, *args, **kwargs)


def try_timer_coro_run_on(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], delay: float, coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[Optional[CoroID]]:
    """_summary_
        context can be generated by one of the [interface_and_loop_with_backup_loop, get_interface_and_loop_with_backup_loop, interface_and_loop_with_explicit_loop, get_interface_and_loop_with_explicit_loop, interface_for_an_explicit_loop, get_interface_for_an_explicit_loop] functions from the cengal/parallel_execution/coroutines/coro_scheduler module
        
        An example:

        from cengal.parallel_execution.coroutines.coro_scheduler import get_interface_and_loop_with_explicit_loop, CoroSchedulerType, ExplicitWorker, Worker, CoroID
        from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import try_put_coro_to
        from typing import Optional, Union

        def my_func(loop: CoroSchedulerType, coro_worker: AnyWorker, a, b) -> Optional[CoroID]:
            def print_hello_world(i: Interface, name: str):
                print(f'Hello Wrold from {name}!)
            
            try_timer_coro_run_on(10, print_hello_world, 'John Doe')
        
    Args:
        context (Tuple[Optional[CoroSchedulerType], Optional[Interface], bool]): _description_
        delay (float): delay in seconds
        coro_worker (AnyWorker): coro_worker

    Returns:
        ValueExistence[Optional[CoroID]]: _description_
    """
    return try_make_request_to_service_with_context(context, TimerCoroRunner, delay, coro_worker, *args, **kwargs)


async def atimer_coro_run_on(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], delay: float, coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[CoroID]:
    return await amake_request_to_service_with_context(context, TimerCoroRunner, delay, coro_worker, *args, **kwargs)


async def atry_timer_coro_run_on(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], delay: float, coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[Optional[CoroID]]:
    return await atry_make_request_to_service_with_context(context, TimerCoroRunner, delay, coro_worker, *args, **kwargs)


def timer_coro_run(delay: float, coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[CoroID]:
    return make_request_to_service(TimerCoroRunner, delay, coro_worker, *args, **kwargs)


def try_timer_coro_run(delay: float, coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[Optional[CoroID]]:
    return try_make_request_to_service(TimerCoroRunner, delay, coro_worker, *args, **kwargs)


async def atimer_coro_run(delay: float, coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[CoroID]:
    return await amake_request_to_service(TimerCoroRunner, delay, coro_worker, *args, **kwargs)


async def atry_timer_coro_run(delay: float, coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[Optional[CoroID]]:
    return await atry_make_request_to_service(TimerCoroRunner, delay, coro_worker, *args, **kwargs)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from threading import current_thread
from cengal.text_processing.text_processing import find_text, normalize_text
from cengal.text_processing.utf_bom_processing import remove_bom, determine_text_bom
from cengal.performance_test_lib.performance_test_lib import test_performance
from array import array

s_p =('qwertyuiop'*102)*25
s0 = 'Hello World!'
s = s_p + s0 + s_p
s_1 = s0 + s_p + s_p
s2 = 'orl'
s3 = 'World!'
s4 = 'Hello'

b_1 = normalize_text(s_1, bytearray, 'utf-32')
b_1 = remove_bom(b_1, determine_text_bom(b_1))
b4 = normalize_text(s4, bytearray, 'utf-32')
b4 = remove_bom(b4, determine_text_bom(b4))
a_1 = array('L', b_1)
a4 = array('L', b4)

print(f'> len(s): {len(s)}')
print(f'> len(s2): {len(s2)}')
s2_place = find_text(s, s2)
print('s2_place:', s2_place)
print('s[s2_place]:', s[s2_place])
b = normalize_text(s, bytearray, 'utf-32')
print(f'> len(b): {len(b)}')
b = remove_bom(b, determine_text_bom(b))
print(f'> len(b): {len(b)}')
b2 = normalize_text(s2, bytearray, 'utf-32')
bom = determine_text_bom(b2)
print(f'> len(b2): {len(b2)}')
print(f'b2: {b2}; len: {len(b2)}')
b2 = remove_bom(b2, bom)

ba = bytearray(b)
ba2 = bytearray(b2)

print(f'> len(b2): {len(b2)}')
print(f'b2: {b2}; len: {len(b2)}')
b2_place = find_text(b, b2)
print('b2_place:', b2_place)
print('b->str slice:', (bom + (b[b2_place])).decode('utf-32'))
print('b2->str:', (bom + b2).decode('utf-32'))
print(b[b2_place] == b2)
m = memoryview(b)
m2 = memoryview(b2)
print(m[b2_place] == m2)

print()
print('itemsize:', array('L').itemsize)
print(f'len b: {len(b)}')
a = array('L', b)
print(f'> len(a): {len(a)}')
a2 = array('L', b2)
print(f'> len(a2): {len(a2)}')
print(f'a2: {a2}')
a2_place = slice(int(b2_place.start / 4), int(b2_place.stop / 4))
ma = memoryview(a)
ma2 = memoryview(a2)
print(ma[a2_place] == ma2)

b2_len = len(b2)
ba2_len = len(ba2)
s2_len = len(s2)
s3_len = len(s3)
s4_len = len(s4)
b4_len = len(b4)
b4_len = len(b4)
a4_len = len(a4)

with test_performance('b[b2_place] == b2', 1) as tracer:
    while tracer.iter():
        b[b2_place] == b2

with test_performance('m[b2_place] == m2', 1) as tracer:
    while tracer.iter():
        m[b2_place] == m2

with test_performance('a[a2_place] == a2', 1) as tracer:
    while tracer.iter():
        a[a2_place] == a2

with test_performance('ma[a2_place] == ma2', 1) as tracer:
    while tracer.iter():
        ma[a2_place] == ma2

with test_performance('bytes(ma)', 1) as tracer:
    while tracer.iter():
        result = bytes(ma)

with test_performance('ma.tobytes()', 1) as tracer:
    while tracer.iter():
        result = ma.tobytes()

with test_performance('a.tobytes()', 1) as tracer:
    while tracer.iter():
        result = a.tobytes()

with test_performance('str.find(s2)', 1) as tracer:
    while tracer.iter():
        start = s.find(s2)
        place = slice(start, start + s2_len)

with test_performance('find_text(s, s2)', 1) as tracer:
    while tracer.iter():
        place = find_text(s, s2)

with test_performance('str.find(s3)', 1) as tracer:
    while tracer.iter():
        start = s.find(s3)
        place = slice(start, start + s3_len)

with test_performance('find_text(s, s3)', 1) as tracer:
    while tracer.iter():
        place = find_text(s, s3)

with test_performance('s_1.startswith(s4)', 1) as tracer:
    while tracer.iter():
        if s_1.startswith(s4):
            place = slice(0, s4_len)
        else:
            place = None

with test_performance('b_1.startswith(b4)', 1) as tracer:
    while tracer.iter():
        if b_1.startswith(b4):
            place = slice(0, b4_len)
        else:
            place = None

with test_performance('a_1[:a4_len] == a4', 1) as tracer:
    while tracer.iter():
        if a_1[:a4_len] == a4:
            place = slice(0, a4_len)
        else:
            place = None

with test_performance('b_1[:b4_len] == a4', 1) as tracer:
    while tracer.iter():
        if b_1[:b4_len] == b4:
            place = slice(0, b4_len)
        else:
            place = None

with test_performance('s_1[:s4_len] == s4', 1) as tracer:
    while tracer.iter():
        if s_1[:s4_len] == s4:
            place = slice(0, s4_len)
        else:
            place = None

with test_performance('s[s2_place] == s2', 1) as tracer:
    while tracer.iter():
        if s[s2_place] == s2:
            place = slice(s2_place)
        else:
            place = None

with test_performance('s[place] == s2', 1) as tracer:
    while tracer.iter():
        place = slice(102028, 102040)
        if s[place] == s2:
            place = slice(place)
        else:
            place = None

with test_performance('s[slice(102028, 102040)] == s2', 1) as tracer:
    while tracer.iter():
        if s[slice(102028, 102040)] == s2:
            place = slice(102028, 102040)
        else:
            place = None

with test_performance('s[25507:25510] == s2', 1) as tracer:
    while tracer.iter():
        if s[25507:25510] == s2:
            place = slice(25507, 25510)
        else:
            place = None

with test_performance('s[place] == s2 ; slice(s2_place.start, s2_place.start + s2_len)', 1) as tracer:
    while tracer.iter():
        place = slice(s2_place.start, s2_place.start + s2_len)
        if s[place] == s2:
            place = slice(place)
        else:
            place = None

with test_performance('s[slice(s2_place.start, s2_place.start + s2_len)] == s2', 1) as tracer:
    while tracer.iter():
        if s[slice(s2_place.start, s2_place.start + s2_len)] == s2:
            place = slice(s2_place.start, s2_place.start + s2_len)
        else:
            place = None

with test_performance('s[place] == s2 ; slice(start, start + s2_len)', 1) as tracer:
    while tracer.iter():
        start = s2_place.start
        place = slice(start, start + s2_len)
        if s[place] == s2:
            place = slice(place)
        else:
            place = None

with test_performance('s[slice(start, end)] == s2', 1) as tracer:
    # Fastest
    while tracer.iter():
        start = s2_place.start
        end = start + s2_len
        if s[slice(start, end)] == s2:
            place = slice(start, end)
        else:
            place = None

with test_performance('bytearray.find', 1) as tracer:
    while tracer.iter():
        start = ba.find(ba2)
        place = slice(start, start + ba2_len)

with test_performance('find_text(b, b2)', 1) as tracer:
    while tracer.iter():
        place = find_text(ba, ba2)

with test_performance('bytes.find', 1) as tracer:
    while tracer.iter():
        start = b.find(b2)
        place = slice(start, start + b2_len)

with test_performance('find_text(b, b2)', 1) as tracer:
    while tracer.iter():
        place = find_text(b, b2)

with test_performance('ma.tobytes().find(b2) /', 1) as tracer:
    while tracer.iter():
        start = ma.tobytes().find(b2)
        place = slice(int(start / 4), int((start + b2_len) / 4))

with test_performance('ma.tobytes().find(b2) >>', 1) as tracer:
    while tracer.iter():
        start = ma.tobytes().find(b2)
        place = slice(start >> 2, (start + b2_len) >> 2)

with test_performance('a.tobytes().find(b2) >>', 1) as tracer:
    while tracer.iter():
        start = a.tobytes().find(b2)
        place = slice(start >> 2, (start + b2_len) >> 2)

with test_performance('find_text(ma.tobytes(), b2)', 1) as tracer:
    while tracer.iter():
        place = find_text(ma.tobytes(), b2)
        place = slice(place.start >> 2, (place.start + b2_len) >> 2)

